"""Unit tests for DiffLayout and DiffFormatContext."""

from __future__ import annotations

import pytest

from rconfig.diff import ConfigDiff, DiffEntry, DiffEntryType, DiffFormat, DiffFormatContext, DiffLayout
from rconfig.diff.formatting.model import DiffDisplayModel, DiffDisplayModelBuilder
from rconfig.provenance.models import ProvenanceEntry


class TestDiffFormatContext:
    """Tests for DiffFormatContext dataclass."""

    def test_default_values(self) -> None:
        """DiffFormatContext has expected default values."""
        ctx = DiffFormatContext()

        assert ctx.show_paths is True
        assert ctx.show_values is True
        assert ctx.show_files is True
        assert ctx.show_lines is True
        assert ctx.show_provenance is False
        assert ctx.show_unchanged is False
        assert ctx.show_added is True
        assert ctx.show_removed is True
        assert ctx.show_changed is True
        assert ctx.show_counts is True
        assert ctx.indent_size == 2
        assert ctx.path_filters == []
        assert ctx.file_filters == []

    def test_custom_values(self) -> None:
        """DiffFormatContext can be created with custom values."""
        ctx = DiffFormatContext(
            show_paths=False,
            show_values=False,
            show_provenance=True,
            show_unchanged=True,
            indent_size=4,
            path_filters=["model.*"],
            file_filters=["*.yaml"],
        )

        assert ctx.show_paths is False
        assert ctx.show_values is False
        assert ctx.show_provenance is True
        assert ctx.show_unchanged is True
        assert ctx.indent_size == 4
        assert ctx.path_filters == ["model.*"]
        assert ctx.file_filters == ["*.yaml"]

    def test_is_mutable(self) -> None:
        """DiffFormatContext fields can be mutated."""
        ctx = DiffFormatContext()
        ctx.show_paths = False
        ctx.path_filters.append("test.*")

        assert ctx.show_paths is False
        assert "test.*" in ctx.path_filters


class TestDiffLayoutABC:
    """Tests for DiffLayout abstract base class."""

    def test_cannot_instantiate_directly(self) -> None:
        """DiffLayout is abstract and cannot be instantiated directly."""
        with pytest.raises(TypeError):
            DiffLayout()  # type: ignore

    def test_concrete_subclass_must_implement_render(self) -> None:
        """Concrete subclass must implement render method."""

        class IncompleteLayout(DiffLayout):
            pass

        with pytest.raises(TypeError):
            IncompleteLayout()  # type: ignore

    def test_concrete_subclass_with_render_works(self) -> None:
        """Concrete subclass with render method can be instantiated."""

        class MinimalLayout(DiffLayout):
            def render(self, model: DiffDisplayModel) -> str:
                return "rendered"

        layout = MinimalLayout()
        assert layout is not None


class TestDiffFormatValueFormatting:
    """Tests for value formatting used by DiffFormat.

    Value formatting is provided by format_value() from format_utils.
    """

    def test_format_value_none(self) -> None:
        """format_value handles None."""
        from rconfig._internal.format_utils import format_value

        assert format_value(None) == "null"

    def test_format_value_bool(self) -> None:
        """format_value handles booleans."""
        from rconfig._internal.format_utils import format_value

        assert format_value(True) == "true"
        assert format_value(False) == "false"

    def test_format_value_string(self) -> None:
        """format_value handles strings with quotes."""
        from rconfig._internal.format_utils import format_value

        assert format_value("hello") == "'hello'"

    def test_format_value_number(self) -> None:
        """format_value handles numbers."""
        from rconfig._internal.format_utils import format_value

        assert format_value(42) == "42"
        assert format_value(3.14) == "3.14"

    def test_format_value_truncates_long_values(self) -> None:
        """format_value truncates long lists/dicts."""
        from rconfig._internal.format_utils import format_value

        long_list = list(range(100))
        result = format_value(long_list)
        assert len(result) <= 50
        assert result.endswith("...")


class TestDiffFormatBuildModel:
    """Tests for DiffFormat._build_model()."""

    def test_build_creates_display_model(self) -> None:
        """_build_model() creates a complete display model."""
        entries = {
            "a": DiffEntry("a", DiffEntryType.ADDED, right_value=1),
            "b": DiffEntry("b", DiffEntryType.REMOVED, left_value=2),
            "c": DiffEntry("c", DiffEntryType.CHANGED, 3, 4),
        }
        diff = ConfigDiff(entries)
        fmt = DiffFormat(diff)

        model = fmt._build_model()

        assert len(model.entries) == 3
        assert model.empty_message is None

        # Check entry types
        added = [e for e in model.entries if e.diff_type == DiffEntryType.ADDED]
        removed = [e for e in model.entries if e.diff_type == DiffEntryType.REMOVED]
        changed = [e for e in model.entries if e.diff_type == DiffEntryType.CHANGED]
        assert len(added) == 1
        assert len(removed) == 1
        assert len(changed) == 1

    def test_build_respects_visibility_flags(self) -> None:
        """_build_model() respects show_added, show_removed, etc. flags."""
        entries = {
            "a": DiffEntry("a", DiffEntryType.ADDED, right_value=1),
            "b": DiffEntry("b", DiffEntryType.REMOVED, left_value=2),
        }
        diff = ConfigDiff(entries)
        fmt = DiffFormat(diff)
        fmt._ctx.show_added = False

        model = fmt._build_model()

        # Added should be filtered out
        added = [e for e in model.entries if e.diff_type == DiffEntryType.ADDED]
        removed = [e for e in model.entries if e.diff_type == DiffEntryType.REMOVED]
        assert len(added) == 0
        assert len(removed) == 1

    def test_build_empty_diff_shows_message(self) -> None:
        """_build_model() sets empty_message for empty diff."""
        diff = ConfigDiff({})
        fmt = DiffFormat(diff)

        model = fmt._build_model()

        assert model.empty_message is not None
        assert "No differences found" in model.empty_message





"""Tests for Pydantic model instantiation.

These tests verify that rconfig correctly instantiates Pydantic BaseModel classes,
including frozen models, nested structures, and implicit target inference.
"""

from typing import Optional
from unittest import TestCase

from pydantic import BaseModel, ConfigDict, ValidationError

from rconfig.target import TargetRegistry
from rconfig.validation import ConfigValidator
from rconfig.instantiation import ConfigInstantiator


class PydanticInstantiationTests(TestCase):
    """Tests for basic Pydantic BaseModel instantiation."""

    def _empty_store(self) -> TargetRegistry:
        store = TargetRegistry()
        store.clear()
        return store

    def _create_instantiator(self, store: TargetRegistry) -> ConfigInstantiator:
        validator = ConfigValidator(store)
        return ConfigInstantiator(store, validator)

    def test_instantiate__PydanticBaseModel__ReturnsInstance(self):
        # Arrange
        store = self._empty_store()

        class Model(BaseModel):
            hidden_size: int
            dropout: float = 0.1

        store.register("model", Model)
        instantiator = self._create_instantiator(store)
        config = {"_target_": "model", "hidden_size": 256}

        # Act
        result = instantiator.instantiate(config)

        # Assert
        self.assertIsInstance(result, Model)
        self.assertEqual(result.hidden_size, 256)
        self.assertEqual(result.dropout, 0.1)

    def test_instantiate__AllFieldsProvided__UsesProvidedValues(self):
        # Arrange
        store = self._empty_store()

        class Model(BaseModel):
            size: int
            name: str
            enabled: bool

        store.register("model", Model)
        instantiator = self._create_instantiator(store)
        config = {
            "_target_": "model",
            "size": 100,
            "name": "test",
            "enabled": True,
        }

        # Act
        result = instantiator.instantiate(config)

        # Assert
        self.assertEqual(result.size, 100)
        self.assertEqual(result.name, "test")
        self.assertTrue(result.enabled)

    def test_instantiate__NestedPydanticModels__InstantiatesRecursively(self):
        # Arrange
        store = self._empty_store()

        class InnerModel(BaseModel):
            value: int

        class OuterModel(BaseModel):
            inner: InnerModel
            name: str

        store.register("inner", InnerModel)
        store.register("outer", OuterModel)
        instantiator = self._create_instantiator(store)
        config = {
            "_target_": "outer",
            "inner": {"_target_": "inner", "value": 42},
            "name": "test",
        }

        # Act
        result = instantiator.instantiate(config)

        # Assert
        self.assertIsInstance(result, OuterModel)
        self.assertIsInstance(result.inner, InnerModel)
        self.assertEqual(result.inner.value, 42)
        self.assertEqual(result.name, "test")

    def test_instantiate__DeeplyNestedPydantic__InstantiatesAllLevels(self):
        # Arrange
        store = self._empty_store()

        class Level3(BaseModel):
            value: int

        class Level2(BaseModel):
            level3: Level3

        class Level1(BaseModel):
            level2: Level2

        store.register("l3", Level3)
        store.register("l2", Level2)
        store.register("l1", Level1)
        instantiator = self._create_instantiator(store)
        config = {
            "_target_": "l1",
            "level2": {
                "_target_": "l2",
                "level3": {"_target_": "l3", "value": 99},
            },
        }

        # Act
        result = instantiator.instantiate(config)

        # Assert
        self.assertIsInstance(result, Level1)
        self.assertIsInstance(result.level2, Level2)
        self.assertIsInstance(result.level2.level3, Level3)
        self.assertEqual(result.level2.level3.value, 99)

    def test_instantiate__PydanticWithDefaults__UsesDefaultValues(self):
        # Arrange
        store = self._empty_store()

        class Model(BaseModel):
            required_field: int
            optional_with_default: str = "default_value"
            number_with_default: float = 3.14

        store.register("model", Model)
        instantiator = self._create_instantiator(store)
        config = {"_target_": "model", "required_field": 42}

        # Act
        result = instantiator.instantiate(config)

        # Assert
        self.assertEqual(result.required_field, 42)
        self.assertEqual(result.optional_with_default, "default_value")
        self.assertEqual(result.number_with_default, 3.14)

    def test_instantiate__PydanticOptionalField__AcceptsNone(self):
        # Arrange
        store = self._empty_store()

        class Model(BaseModel):
            required: int
            optional: Optional[str] = None

        store.register("model", Model)
        instantiator = self._create_instantiator(store)
        config = {"_target_": "model", "required": 10, "optional": None}

        # Act
        result = instantiator.instantiate(config)

        # Assert
        self.assertEqual(result.required, 10)
        self.assertIsNone(result.optional)


class PydanticFrozenModelTests(TestCase):
    """Tests for frozen Pydantic model instantiation and immutability."""

    def _empty_store(self) -> TargetRegistry:
        store = TargetRegistry()
        store.clear()
        return store

    def _create_instantiator(self, store: TargetRegistry) -> ConfigInstantiator:
        validator = ConfigValidator(store)
        return ConfigInstantiator(store, validator)

    def test_instantiate__FrozenPydanticModel__ReturnsImmutableInstance(self):
        """Test that frozen Pydantic models are instantiated correctly."""
        # Arrange
        store = self._empty_store()

        class FrozenModel(BaseModel):
            model_config = ConfigDict(frozen=True)

            hidden_size: int
            dropout: float = 0.1

        store.register("frozen_model", FrozenModel)
        instantiator = self._create_instantiator(store)
        config = {"_target_": "frozen_model", "hidden_size": 256}

        # Act
        result = instantiator.instantiate(config)

        # Assert
        self.assertIsInstance(result, FrozenModel)
        self.assertEqual(result.hidden_size, 256)
        self.assertEqual(result.dropout, 0.1)

    def test_instantiate__FrozenPydanticModel__ModificationRaisesValidationError(self):
        """Test that attempting to modify a frozen Pydantic model raises ValidationError."""
        # Arrange
        store = self._empty_store()

        class FrozenModel(BaseModel):
            model_config = ConfigDict(frozen=True)

            value: int

        store.register("frozen_model", FrozenModel)
        instantiator = self._create_instantiator(store)
        config = {"_target_": "frozen_model", "value": 42}
        result = instantiator.instantiate(config)

        # Act & Assert
        with self.assertRaises(ValidationError):
            result.value = 100

    def test_instantiate__NestedFrozenPydantic__BothAreImmutable(self):
        """Test that nested frozen Pydantic models maintain immutability."""
        # Arrange
        store = self._empty_store()

        class InnerFrozen(BaseModel):
            model_config = ConfigDict(frozen=True)

            value: int

        class OuterFrozen(BaseModel):
            model_config = ConfigDict(frozen=True)

            inner: InnerFrozen
            name: str

        store.register("inner_frozen", InnerFrozen)
        store.register("outer_frozen", OuterFrozen)
        instantiator = self._create_instantiator(store)
        config = {
            "_target_": "outer_frozen",
            "inner": {"_target_": "inner_frozen", "value": 42},
            "name": "test",
        }

        # Act
        result = instantiator.instantiate(config)

        # Assert
        self.assertIsInstance(result, OuterFrozen)
        self.assertIsInstance(result.inner, InnerFrozen)
        with self.assertRaises(ValidationError):
            result.name = "modified"
        with self.assertRaises(ValidationError):
            result.inner.value = 100


class PydanticImplicitTargetTests(TestCase):
    """Tests for implicit target inference with Pydantic models."""

    def _empty_store(self) -> TargetRegistry:
        store = TargetRegistry()
        store.clear()
        return store

    def _create_instantiator(self, store: TargetRegistry) -> ConfigInstantiator:
        validator = ConfigValidator(store)
        return ConfigInstantiator(store, validator)

    def test_instantiate__ImplicitNestedPydantic__InfersTarget(self):
        """Test that nested Pydantic models can infer _target_ from type hints."""
        # Arrange
        store = self._empty_store()

        class InnerModel(BaseModel):
            value: int

        class OuterModel(BaseModel):
            inner: InnerModel
            name: str

        store.register("inner", InnerModel)
        store.register("outer", OuterModel)
        instantiator = self._create_instantiator(store)

        # Config without explicit _target_ for inner
        config = {
            "_target_": "outer",
            "inner": {"value": 42},  # No _target_ - should be inferred
            "name": "test",
        }

        # Act
        result = instantiator.instantiate(config)

        # Assert
        self.assertIsInstance(result, OuterModel)
        self.assertIsInstance(result.inner, InnerModel)
        self.assertEqual(result.inner.value, 42)

    def test_instantiate__DeeplyNestedImplicitPydantic__InfersAllTargets(self):
        """Test deeply nested Pydantic models with implicit targets."""
        # Arrange
        store = self._empty_store()

        class Level3(BaseModel):
            value: int

        class Level2(BaseModel):
            level3: Level3

        class Level1(BaseModel):
            level2: Level2

        store.register("l3", Level3)
        store.register("l2", Level2)
        store.register("l1", Level1)
        instantiator = self._create_instantiator(store)

        # Config with implicit targets at all nested levels
        config = {
            "_target_": "l1",
            "level2": {
                "level3": {"value": 99},
            },
        }

        # Act
        result = instantiator.instantiate(config)

        # Assert
        self.assertIsInstance(result, Level1)
        self.assertIsInstance(result.level2, Level2)
        self.assertIsInstance(result.level2.level3, Level3)
        self.assertEqual(result.level2.level3.value, 99)

    def test_instantiate__MixedExplicitImplicitPydantic__ReturnsCorrectInstance(self):
        """Test Pydantic models with both explicit and implicit targets."""
        # Arrange
        store = self._empty_store()

        class Config1(BaseModel):
            value: int

        class Config2(BaseModel):
            value: str

        class Container(BaseModel):
            config1: Config1
            config2: Config2

        store.register("c1", Config1)
        store.register("c2", Config2)
        store.register("container", Container)
        instantiator = self._create_instantiator(store)

        config = {
            "_target_": "container",
            "config1": {"_target_": "c1", "value": 42},  # Explicit
            "config2": {"value": "hello"},  # Implicit
        }

        # Act
        result = instantiator.instantiate(config)

        # Assert
        self.assertIsInstance(result, Container)
        self.assertIsInstance(result.config1, Config1)
        self.assertIsInstance(result.config2, Config2)
        self.assertEqual(result.config1.value, 42)
        self.assertEqual(result.config2.value, "hello")


class PydanticListTests(TestCase):
    """Tests for Pydantic models with list fields."""

    def _empty_store(self) -> TargetRegistry:
        store = TargetRegistry()
        store.clear()
        return store

    def _create_instantiator(self, store: TargetRegistry) -> ConfigInstantiator:
        validator = ConfigValidator(store)
        return ConfigInstantiator(store, validator)

    def test_instantiate__ListOfPydanticModels__InstantiatesAll(self):
        """Test instantiation of list containing Pydantic models."""
        # Arrange
        store = self._empty_store()

        class Item(BaseModel):
            value: int

        class Container(BaseModel):
            items: list[Item]

        store.register("item", Item)
        store.register("container", Container)
        instantiator = self._create_instantiator(store)
        config = {
            "_target_": "container",
            "items": [
                {"_target_": "item", "value": 1},
                {"_target_": "item", "value": 2},
                {"_target_": "item", "value": 3},
            ],
        }

        # Act
        result = instantiator.instantiate(config)

        # Assert
        self.assertIsInstance(result, Container)
        self.assertEqual(len(result.items), 3)
        for i, item in enumerate(result.items, 1):
            self.assertIsInstance(item, Item)
            self.assertEqual(item.value, i)



"""Shared formatting utilities for provenance and diff layouts.

This module provides common formatting functions used across
multiple layout implementations to eliminate code duplication.
"""

from __future__ import annotations

from typing import Any


def format_value(value: Any, max_length: int = 50) -> str:
    """Format a value for display.

    Handles None, bool, str, and collection types with
    appropriate formatting and truncation.

    :param value: The value to format.
    :param max_length: Maximum string length before truncation.
    :return: Formatted value string.
    """
    match value:
        case None:
            return "null"
        case bool():
            return "true" if value else "false"
        case str():
            return repr(value)
        case list() | dict():
            s = str(value)
            if len(s) > max_length:
                return s[: max_length - 3] + "..."
            return s
        case _:
            return str(value)


def format_location(file: str | None, line: int | None) -> str:
    """Format file:line location string.

    :param file: The source file name.
    :param line: The line number, or None.
    :return: Formatted location string, or empty string if no file.
    """
    if not file:
        return ""
    if line is not None:
        return f"{file}:{line}"
    return file


def indent(text: str, level: int, indent_size: int = 2) -> str:
    """Add indentation to text.

    :param text: Text to indent.
    :param level: Indentation level.
    :param indent_size: Number of spaces per indentation level.
    :return: Indented text.
    """
    return " " * (level * indent_size) + text


def format_type_hint(type_hint: Any, prefix: str = "Type") -> str:
    """Format a type hint for display.

    :param type_hint: The type hint to format.
    :param prefix: Prefix for the formatted string (e.g., "Type" or "type").
    :return: Formatted type hint string, or empty string if None.
    """
    if type_hint is None:
        return ""
    if hasattr(type_hint, "__name__"):
        return f"{prefix}: {type_hint.__name__}"
    return f"{prefix}: {type_hint}"



"""Unit tests for inner_path parameter in instantiate_multirun.

These tests verify that inner_path works correctly to instantiate
only a subsection of the config during multirun sweeps.
"""

from dataclasses import dataclass
from pathlib import Path
from unittest import TestCase

import rconfig as rc
from rconfig.composition import clear_cache
from rconfig.errors import InvalidInnerPathError

from rconfig_unittests.fixtures import MockFileSystem, mock_filesystem


@dataclass
class ModelConfig:
    """Test model with configurable parameters."""

    hidden_size: int
    lr: float = 0.001


@dataclass
class EncoderConfig:
    """Nested encoder for testing deep paths."""

    hidden_dim: int
    layers: int = 4


@dataclass
class NestedModelConfig:
    """Model with nested encoder."""

    encoder: EncoderConfig
    dropout: float = 0.1


@dataclass
class TrainerConfig:
    """Full trainer config with model."""

    model: ModelConfig
    epochs: int


class InnerPathMultirunTests(TestCase):
    """Tests for inner_path parameter in instantiate_multirun."""

    def setUp(self):
        rc._store.clear()
        clear_cache()
        rc.register("model", ModelConfig)
        rc.register("trainer", TrainerConfig)

    def test_instantiate_multirun__inner_path_selects_subsection__returns_instantiated_subsection(
        self,
    ):
        """Verify inner_path extracts and instantiates only the specified section."""
        # Arrange
        fs = MockFileSystem("/configs")
        fs.add_file(
            "/configs/trainer.yaml",
            """
_target_: trainer
model:
  _target_: model
  hidden_size: 256
  lr: 0.001
epochs: 10
""",
        )

        with mock_filesystem(fs):
            # Act - sweep path is relative to full config, not inner_path
            results = list(
                rc.instantiate_multirun(
                    path=Path("/configs/trainer.yaml"),
                    inner_path="model",
                    sweep={"model.lr": [0.01]},
                    cli_overrides=False,
                )
            )

            # Assert
            self.assertEqual(len(results), 1)
            # The result should be a ModelConfig, not TrainerConfig
            self.assertIsInstance(results[0].instance, ModelConfig)
            self.assertEqual(results[0].instance.hidden_size, 256)
            self.assertEqual(results[0].instance.lr, 0.01)

    def test_instantiate_multirun__inner_path_with_sweep__applies_sweep_to_subsection(
        self,
    ):
        """Verify sweep parameters work correctly with inner_path."""
        # Arrange
        fs = MockFileSystem("/configs")
        fs.add_file(
            "/configs/trainer.yaml",
            """
_target_: trainer
model:
  _target_: model
  hidden_size: 256
  lr: 0.001
epochs: 10
""",
        )

        with mock_filesystem(fs):
            # Act - sweep path is relative to full config
            results = list(
                rc.instantiate_multirun(
                    path=Path("/configs/trainer.yaml"),
                    inner_path="model",
                    sweep={"model.lr": [0.01, 0.001]},
                    cli_overrides=False,
                )
            )

            # Assert
            self.assertEqual(len(results), 2)
            learning_rates = [r.instance.lr for r in results]
            self.assertIn(0.01, learning_rates)
            self.assertIn(0.001, learning_rates)

    def test_instantiate_multirun__inner_path_invalid_path__raises_InvalidInnerPathError(
        self,
    ):
        """Verify invalid inner_path raises appropriate error."""
        # Arrange
        fs = MockFileSystem("/configs")
        fs.add_file(
            "/configs/trainer.yaml",
            """
_target_: trainer
model:
  _target_: model
  hidden_size: 256
epochs: 10
""",
        )

        with mock_filesystem(fs):
            # Act - use a valid sweep path, but invalid inner_path
            results = rc.instantiate_multirun(
                path=Path("/configs/trainer.yaml"),
                inner_path="nonexistent",
                sweep={"epochs": [5]},
                cli_overrides=False,
            )

            # Assert - error is raised when accessing instance
            result = next(iter(results))
            with self.assertRaises(InvalidInnerPathError):
                _ = result.instance

    def test_instantiate_multirun__inner_path_with_experiments__works_correctly(self):
        """Verify inner_path works with experiments parameter."""
        # Arrange
        fs = MockFileSystem("/configs")
        fs.add_file(
            "/configs/trainer.yaml",
            """
_target_: trainer
model:
  _target_: model
  hidden_size: 256
  lr: 0.001
epochs: 10
""",
        )

        with mock_filesystem(fs):
            # Act - experiment paths are relative to full config
            results = list(
                rc.instantiate_multirun(
                    path=Path("/configs/trainer.yaml"),
                    inner_path="model",
                    experiments=[{"model.hidden_size": 512}, {"model.hidden_size": 1024}],
                    cli_overrides=False,
                )
            )

            # Assert
            self.assertEqual(len(results), 2)
            hidden_sizes = [r.instance.hidden_size for r in results]
            self.assertIn(512, hidden_sizes)
            self.assertIn(1024, hidden_sizes)

    def test_instantiate_multirun__inner_path_with_overrides__applies_to_subsection(
        self,
    ):
        """Verify constant overrides work with inner_path."""
        # Arrange
        fs = MockFileSystem("/configs")
        fs.add_file(
            "/configs/trainer.yaml",
            """
_target_: trainer
model:
  _target_: model
  hidden_size: 256
  lr: 0.001
epochs: 10
""",
        )

        with mock_filesystem(fs):
            # Act - all paths are relative to full config
            results = list(
                rc.instantiate_multirun(
                    path=Path("/configs/trainer.yaml"),
                    inner_path="model",
                    sweep={"model.lr": [0.01, 0.001]},
                    overrides={"model.hidden_size": 512},
                    cli_overrides=False,
                )
            )

            # Assert
            self.assertEqual(len(results), 2)
            # All results should have the overridden hidden_size
            for result in results:
                self.assertEqual(result.instance.hidden_size, 512)

    def test_instantiate_multirun__inner_path_none__full_config_instantiated(self):
        """Verify None inner_path (default) instantiates full config."""
        # Arrange
        fs = MockFileSystem("/configs")
        fs.add_file(
            "/configs/trainer.yaml",
            """
_target_: trainer
model:
  _target_: model
  hidden_size: 256
epochs: 10
""",
        )

        with mock_filesystem(fs):
            # Act
            results = list(
                rc.instantiate_multirun(
                    path=Path("/configs/trainer.yaml"),
                    inner_path=None,  # Explicit None
                    sweep={"epochs": [5, 10]},
                    cli_overrides=False,
                )
            )

            # Assert
            self.assertEqual(len(results), 2)
            # Should be TrainerConfig, not ModelConfig
            self.assertIsInstance(results[0].instance, TrainerConfig)
            self.assertIsInstance(results[0].instance.model, ModelConfig)


class NestedInnerPathTests(TestCase):
    """Tests for nested inner_path like 'model.encoder'."""

    def setUp(self):
        rc._store.clear()
        clear_cache()
        rc.register("encoder", EncoderConfig)
        rc.register("nested_model", NestedModelConfig)

    def test_instantiate_multirun__inner_path_nested__extracts_deeply_nested_section(
        self,
    ):
        """Verify nested paths like 'model.encoder' work."""
        # Arrange
        fs = MockFileSystem("/configs")
        fs.add_file(
            "/configs/model.yaml",
            """
_target_: nested_model
encoder:
  _target_: encoder
  hidden_dim: 256
  layers: 4
dropout: 0.1
""",
        )

        with mock_filesystem(fs):
            # Act - sweep path is relative to full config
            results = list(
                rc.instantiate_multirun(
                    path=Path("/configs/model.yaml"),
                    inner_path="encoder",
                    sweep={"encoder.hidden_dim": [256, 512]},
                    cli_overrides=False,
                )
            )

            # Assert
            self.assertEqual(len(results), 2)
            # Should be EncoderConfig instances
            for result in results:
                self.assertIsInstance(result.instance, EncoderConfig)
            hidden_dims = [r.instance.hidden_dim for r in results]
            self.assertIn(256, hidden_dims)
            self.assertIn(512, hidden_dims)


class InnerPathWithExternalInstanceTests(TestCase):
    """Tests for inner_path with _instance_ references outside the inner path."""

    def setUp(self):
        rc._store.clear()
        clear_cache()

    def test_instantiate_multirun__inner_path_with_external_instance_ref__pre_instantiates_external(
        self,
    ):
        """Verify _instance_ refs outside inner_path are pre-instantiated."""

        @dataclass
        class SharedCache:
            name: str

        @dataclass
        class Service:
            cache: SharedCache

        @dataclass
        class App:
            shared_cache: SharedCache
            service: Service

        rc.register("cache", SharedCache)
        rc.register("service", Service)
        rc.register("app", App)

        fs = MockFileSystem("/configs")
        fs.add_file(
            "/configs/app.yaml",
            """
_target_: app
shared_cache:
  _target_: cache
  name: redis
service:
  _target_: service
  cache:
    _instance_: shared_cache
""",
        )

        with mock_filesystem(fs):
            # Act - use valid path for experiment
            results = list(
                rc.instantiate_multirun(
                    path=Path("/configs/app.yaml"),
                    inner_path="service",
                    experiments=[{"shared_cache.name": "memcached"}],
                    cli_overrides=False,
                )
            )

            # Assert
            self.assertEqual(len(results), 1)
            # Service should be instantiated with the shared cache
            self.assertIsInstance(results[0].instance, Service)
            self.assertIsInstance(results[0].instance.cache, SharedCache)


class InnerPathWithSweepAndExperimentsTests(TestCase):
    """Tests combining inner_path with both sweep and experiments."""

    def setUp(self):
        rc._store.clear()
        clear_cache()
        rc.register("model", ModelConfig)
        rc.register("trainer", TrainerConfig)

    def test_instantiate_multirun__inner_path_with_both_sweep_and_experiments__cartesian_product(
        self,
    ):
        """Verify inner_path works with both sweep and experiments (cartesian product)."""
        # Arrange
        fs = MockFileSystem("/configs")
        fs.add_file(
            "/configs/trainer.yaml",
            """
_target_: trainer
model:
  _target_: model
  hidden_size: 256
  lr: 0.001
epochs: 10
""",
        )

        with mock_filesystem(fs):
            # Act - all paths are relative to full config
            results = list(
                rc.instantiate_multirun(
                    path=Path("/configs/trainer.yaml"),
                    inner_path="model",
                    experiments=[{"model.hidden_size": 256}, {"model.hidden_size": 512}],
                    sweep={"model.lr": [0.01, 0.001]},
                    cli_overrides=False,
                )
            )

            # Assert - 2 experiments x 2 lr values = 4 total
            self.assertEqual(len(results), 4)
            for result in results:
                self.assertIsInstance(result.instance, ModelConfig)



"""MultirunResult dataclass for storing instantiation results.

This module defines the result type returned by multirun iterations,
containing the resolved config, applied overrides, and the instantiated object.
"""

from dataclasses import dataclass, field
from types import MappingProxyType
from typing import Any, Generic, TypeVar

T = TypeVar("T")


@dataclass(frozen=True, kw_only=True)
class MultirunResult(Generic[T]):
    """Result of a single multirun instantiation.

    Contains the immutable resolved config, the specific overrides applied
    for this run, and the instantiated object (or stored error).

    The `instance` property raises any stored error when accessed, enabling
    try/except error handling patterns.

    :param config: Immutable view of the fully resolved configuration dict.
    :param overrides: The specific overrides applied for this run.

    Example::

        # Pattern 1: Fail fast (let it raise)
        for result in rc.instantiate_multirun(...):
            train(result.instance)  # Raises if this run failed

        # Pattern 2: Handle errors individually with try/except
        for result in rc.instantiate_multirun(...):
            try:
                train(result.instance)
            except InstantiationError as e:
                log_failure(result.overrides, e)
                continue
    """

    config: MappingProxyType[str, Any]
    overrides: MappingProxyType[str, Any]
    _instance: T | None = field(repr=False)
    _error: Exception | None = field(repr=False, default=None)

    @property
    def instance(self) -> T:
        """Access the instantiated object.

        :return: The instantiated object for this run.
        :raises: The stored error if instantiation failed.
        """
        if self._error is not None:
            raise self._error
        return self._instance  # type: ignore[return-value]



"""Unit tests for multirun error classes."""

from unittest import TestCase

from rconfig.multirun import (
    InvalidSweepValueError,
    NoRunConfigurationError,
    MultirunError,
)
from rconfig.errors import ConfigError


class MultirunErrorTests(TestCase):
    """Tests for MultirunError base class."""

    def test_MultirunError__InheritsFromConfigError(self):
        # Assert
        self.assertTrue(issubclass(MultirunError, ConfigError))


class InvalidSweepValueErrorTests(TestCase):
    """Tests for InvalidSweepValueError."""

    def test_InvalidSweepValueError__Message__ContainsPathAndHint(self):
        # Arrange
        error = InvalidSweepValueError(
            path="callbacks",
            expected="list[list[...]]",
            actual="str",
            index=0,
        )

        # Act
        message = str(error)

        # Assert
        self.assertIn("callbacks", message)
        self.assertIn("index 0", message)
        self.assertIn("list[list[...]]", message)
        self.assertIn("str", message)
        self.assertIn("Hint", message)

    def test_InvalidSweepValueError__AttributesSet(self):
        # Arrange
        error = InvalidSweepValueError(
            path="model.layers",
            expected="list[list[int]]",
            actual="int",
            index=2,
        )

        # Assert
        self.assertEqual(error.path, "model.layers")
        self.assertEqual(error.expected, "list[list[int]]")
        self.assertEqual(error.actual, "int")
        self.assertEqual(error.index, 2)


class NoRunConfigurationErrorTests(TestCase):
    """Tests for NoRunConfigurationError."""

    def test_NoRunConfigurationError__WithOverrides__DescriptiveMessage(self):
        # Arrange
        error = NoRunConfigurationError(has_overrides=True)

        # Act
        message = str(error)

        # Assert
        self.assertIn("No run configuration", message)
        self.assertIn("overrides", message.lower())
        self.assertIn("instantiate()", message)

    def test_NoRunConfigurationError__WithoutOverrides__DescriptiveMessage(self):
        # Arrange
        error = NoRunConfigurationError(has_overrides=False)

        # Act
        message = str(error)

        # Assert
        self.assertIn("No run configuration", message)
        self.assertIn("sweep", message.lower())
        self.assertIn("experiments", message.lower())

    def test_NoRunConfigurationError__AttributeSet(self):
        # Arrange
        error = NoRunConfigurationError(has_overrides=True)

        # Assert
        self.assertTrue(error.has_overrides)



"""Integration tests for deprecation warnings feature."""

import tempfile
import warnings
from pathlib import Path
from unittest import TestCase

import rconfig as rc
from rconfig.deprecation.handler import RconfigDeprecationWarning


class DeprecationIntegrationTests(TestCase):
    """Integration tests for the complete deprecation workflow."""

    def setUp(self) -> None:
        self.temp_dir = tempfile.TemporaryDirectory()
        self.config_root = Path(self.temp_dir.name)
        rc.clear_cache()
        # Clear deprecation registry
        from rconfig.deprecation.registry import get_deprecation_registry

        self.registry = get_deprecation_registry()
        self.registry.clear()

    def tearDown(self) -> None:
        self.temp_dir.cleanup()
        rc.clear_cache()
        self.registry.clear()

    def _write_config(self, name: str, content: str) -> Path:
        """Write a config file to the temp directory."""
        path = self.config_root / name
        path.parent.mkdir(parents=True, exist_ok=True)
        path.write_text(content)
        return path

    def test_deprecate__RegisteredKey__WarningEmitted(self):
        # Arrange
        config_path = self._write_config(
            "config.yaml",
            """
learning_rate: 0.001
epochs: 100
""",
        )
        rc.deprecate("learning_rate", new_key="optimizer.lr")

        # Act
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            prov = rc.get_provenance(config_path)

        # Assert - warning should be emitted (if detection is hooked in)
        # Note: Full integration requires Walker hook which we haven't implemented
        # For now, verify the provenance was returned
        self.assertIsNotNone(prov)

    def test_deprecate__PublicAPI__WorksCorrectly(self):
        # Arrange
        rc.deprecate("old_key", new_key="new_key", message="Test msg", remove_in="2.0")

        # Act
        info = self.registry.get("old_key")

        # Assert
        self.assertIsNotNone(info)
        self.assertEqual(info.pattern, "old_key")
        self.assertEqual(info.new_key, "new_key")
        self.assertEqual(info.message, "Test msg")
        self.assertEqual(info.remove_in, "2.0")

    def test_undeprecate__AfterRegister__RemovesDeprecation(self):
        # Arrange
        rc.deprecate("old_key")

        # Act
        rc.undeprecate("old_key")

        # Assert
        self.assertFalse(self.registry.is_deprecated("old_key"))

    def test_setDeprecationPolicy__ChangesGlobalPolicy(self):
        # Act
        rc.set_deprecation_policy("error")

        # Assert
        self.assertEqual(self.registry.global_policy, "error")

    def test_setDeprecationPolicy__Warn__EmitsWarning(self):
        # Arrange
        rc.set_deprecation_policy("warn")
        rc.deprecate("test_key")

        # Act
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            from rconfig.deprecation.detector import check_deprecation
            from rconfig.provenance import ProvenanceBuilder

            builder = ProvenanceBuilder()
            builder.add("test_key", file="test.yaml", line=1)
            check_deprecation("test_key", "test.yaml", 1, builder)

        # Assert
        self.assertEqual(len(w), 1)
        self.assertTrue(issubclass(w[0].category, RconfigDeprecationWarning))

    def test_setDeprecationPolicy__Error__RaisesException(self):
        # Arrange
        rc.set_deprecation_policy("error")
        rc.deprecate("critical_key")

        # Act & Assert
        with self.assertRaises(rc.DeprecatedKeyError):
            from rconfig.deprecation.detector import check_deprecation
            from rconfig.provenance import ProvenanceBuilder

            builder = ProvenanceBuilder()
            builder.add("critical_key", file="test.yaml", line=1)
            check_deprecation("critical_key", "test.yaml", 1, builder)

    def test_setDeprecationPolicy__Ignore__NoWarningOrError(self):
        # Arrange
        rc.set_deprecation_policy("ignore")
        rc.deprecate("ignored_key")

        # Act
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            from rconfig.deprecation.detector import check_deprecation
            from rconfig.provenance import ProvenanceBuilder

            builder = ProvenanceBuilder()
            builder.add("ignored_key", file="test.yaml", line=1)
            check_deprecation("ignored_key", "test.yaml", 1, builder)

        # Assert - no warnings
        self.assertEqual(len(w), 0)

    def test_deprecationHandler__Decorator__RegistersHandler(self):
        # Arrange
        calls = []

        @rc.deprecation_handler
        def test_handler(info, path, file, line):
            calls.append((path, file, line))

        rc.deprecate("test_key")

        # Act
        from rconfig.deprecation.detector import check_deprecation
        from rconfig.provenance import ProvenanceBuilder

        builder = ProvenanceBuilder()
        builder.add("test_key", file="test.yaml", line=5)
        check_deprecation("test_key", "test.yaml", 5, builder)

        # Assert
        self.assertEqual(len(calls), 1)
        self.assertEqual(calls[0], ("test_key", "test.yaml", 5))

    def test_setDeprecationHandler__CustomHandler__UsesCustom(self):
        # Arrange
        from rconfig.deprecation import DeprecationHandler

        calls = []

        class TestHandler(DeprecationHandler):
            def handle(self, info, path, file, line):
                calls.append((info.pattern, path))

        rc.set_deprecation_handler(TestHandler())
        rc.deprecate("custom_key")

        # Act
        from rconfig.deprecation.detector import check_deprecation
        from rconfig.provenance import ProvenanceBuilder

        builder = ProvenanceBuilder()
        builder.add("custom_key", file="test.yaml", line=1)
        check_deprecation("custom_key", "test.yaml", 1, builder)

        # Assert
        self.assertEqual(len(calls), 1)
        self.assertEqual(calls[0], ("custom_key", "custom_key"))

    def test_deprecationsPreset__FullWorkflow__FormatsCorrectly(self):
        # Arrange
        from rconfig.provenance import ProvenanceBuilder
        from rconfig.deprecation.info import DeprecationInfo

        builder = ProvenanceBuilder()
        builder.add(
            "old_key",
            file="config.yaml",
            line=5,
            value=42,
            deprecation=DeprecationInfo(
                pattern="old_key",
                matched_path="old_key",
                new_key="new_key",
                message="Use new_key instead",
                remove_in="2.0.0",
            ),
        )
        builder.add("normal_key", file="config.yaml", line=10, value="normal")
        prov = builder.build()

        # Act
        result = str(rc.format(prov).deprecations())

        # Assert
        self.assertIn("Deprecated Keys:", result)
        self.assertIn("old_key", result)
        self.assertIn("DEPRECATED", result)
        self.assertIn("new_key", result)
        self.assertIn("2.0.0", result)
        self.assertNotIn("normal_key", result)

    def test_globPattern__DoubleWildcard__MatchesAnyDepth(self):
        # Arrange
        rc.deprecate("**.dropout", message="Dropout configured elsewhere")

        # Act
        from rconfig.deprecation.detector import check_deprecation
        from rconfig.provenance import ProvenanceBuilder

        builder = ProvenanceBuilder()
        builder.add("model.encoder.dropout", file="config.yaml", line=1)

        with warnings.catch_warnings(record=True):
            warnings.simplefilter("always")
            result = check_deprecation(
                "model.encoder.dropout", "config.yaml", 1, builder
            )

        # Assert
        self.assertIsNotNone(result)
        self.assertEqual(result.pattern, "**.dropout")

    def test_globPattern__SingleWildcard__MatchesOneLevel(self):
        # Arrange
        rc.deprecate("*.lr", message="Use full path")

        # Act
        from rconfig.deprecation.detector import check_deprecation
        from rconfig.provenance import ProvenanceBuilder

        # Should match
        builder1 = ProvenanceBuilder()
        builder1.add("model.lr", file="config.yaml", line=1)
        with warnings.catch_warnings(record=True):
            warnings.simplefilter("always")
            result1 = check_deprecation("model.lr", "config.yaml", 1, builder1)

        # Should NOT match (multiple levels)
        builder2 = ProvenanceBuilder()
        builder2.add("model.encoder.lr", file="config.yaml", line=2)
        result2 = check_deprecation("model.encoder.lr", "config.yaml", 2, builder2)

        # Assert
        self.assertIsNotNone(result1)
        self.assertIsNone(result2)

    def test_autoMap__WithNewKey__CreatesIntermediateStructures(self):
        # Arrange
        from rconfig.deprecation.detector import auto_map_deprecated_values
        from rconfig.provenance import ProvenanceBuilder
        from rconfig.deprecation.info import DeprecationInfo

        config = {"learning_rate": 0.001}
        builder = ProvenanceBuilder()
        builder.add(
            "learning_rate",
            file="config.yaml",
            line=1,
            deprecation=DeprecationInfo(
                pattern="learning_rate",
                matched_path="learning_rate",
                new_key="model.optimizer.lr",
            ),
        )

        # Act
        result = auto_map_deprecated_values(config, builder)

        # Assert
        self.assertEqual(result["model"]["optimizer"]["lr"], 0.001)
        self.assertEqual(result["learning_rate"], 0.001)  # Old key preserved

    def test_autoMap__NewKeyExists__DoesNotOverride(self):
        # Arrange
        from rconfig.deprecation.detector import auto_map_deprecated_values
        from rconfig.provenance import ProvenanceBuilder
        from rconfig.deprecation.info import DeprecationInfo

        config = {
            "old_lr": 0.001,
            "new_lr": 0.01,  # Already exists
        }
        builder = ProvenanceBuilder()
        builder.add(
            "old_lr",
            file="config.yaml",
            line=1,
            deprecation=DeprecationInfo(
                pattern="old_lr",
                matched_path="old_lr",
                new_key="new_lr",
            ),
        )

        # Act
        result = auto_map_deprecated_values(config, builder)

        # Assert
        self.assertEqual(result["new_lr"], 0.01)  # Keeps existing value

    def test_perDeprecationPolicy__Override__TakesPrecedence(self):
        # Arrange
        rc.set_deprecation_policy("warn")
        rc.deprecate("critical_key", policy="error")

        # Act & Assert
        with self.assertRaises(rc.DeprecatedKeyError):
            from rconfig.deprecation.detector import check_deprecation
            from rconfig.provenance import ProvenanceBuilder

            builder = ProvenanceBuilder()
            builder.add("critical_key", file="test.yaml", line=1)
            check_deprecation("critical_key", "test.yaml", 1, builder)



"""Integration tests for formatting preset registries."""

from __future__ import annotations

from pathlib import Path
from textwrap import dedent

import pytest

import rconfig as rc
from rconfig.provenance.formatting import get_provenance_registry
from rconfig.diff.formatting import get_diff_registry


class TestProvenancePresetIntegration:
    """Integration tests for provenance preset registration and usage."""

    @pytest.fixture(autouse=True)
    def clear_registries(self):
        """Clear custom presets before and after each test."""
        get_provenance_registry().clear_presets()
        get_diff_registry().clear_presets()
        yield
        get_provenance_registry().clear_presets()
        get_diff_registry().clear_presets()

    def test_ProvenanceFormat__StringPreset__AppliesCorrectly(self, tmp_path: Path) -> None:
        """Test that string-based preset() method applies presets correctly."""
        config_file = tmp_path / "config.yaml"
        config_file.write_text(
            dedent(
                """
                model:
                  lr: 0.01
                  layers: 4
                """
            )
        )

        prov = rc.get_provenance(config_file)

        # Test various presets work
        result_minimal = str(rc.format(prov).preset("minimal"))
        result_full = str(rc.format(prov).preset("full"))
        result_values = str(rc.format(prov).preset("values"))

        # Minimal should have fewer details than full
        assert len(result_minimal) < len(result_full)
        # Values should not contain file info
        assert "config.yaml" not in result_values

    def test_ProvenanceFormat__NamedMethod__AppliesCorrectly(self, tmp_path: Path) -> None:
        """Test that named preset methods work correctly."""
        config_file = tmp_path / "config.yaml"
        config_file.write_text(
            dedent(
                """
                model:
                  lr: 0.01
                """
            )
        )

        prov = rc.get_provenance(config_file)

        # Named methods should produce same output as preset()
        result_minimal_method = str(rc.format(prov).minimal())
        result_minimal_preset = str(rc.format(prov).preset("minimal"))

        assert result_minimal_method == result_minimal_preset

    def test_ProvenanceFormat__CustomPreset__AppliesCorrectly(self, tmp_path: Path) -> None:
        """Test that custom presets can be registered and used."""
        config_file = tmp_path / "config.yaml"
        config_file.write_text(
            dedent(
                """
                model:
                  lr: 0.01
                """
            )
        )

        # Register custom preset
        rc.register_provenance_preset(
            "paths_only",
            lambda: rc.ProvenanceFormatContext(
                show_paths=True,
                show_values=False,
                show_files=False,
                show_lines=False,
                show_source_type=False,
                show_chain=False,
                show_overrides=False,
                show_targets=False,
            ),
            "Show only paths",
        )

        prov = rc.get_provenance(config_file)
        result = str(rc.format(prov).preset("paths_only"))

        # Should contain path but not value
        assert "model.lr" in result
        assert "0.01" not in result

    def test_ProvenanceFormat__UnknownPreset__RaisesValueError(self, tmp_path: Path) -> None:
        """Test that unknown preset raises ValueError."""
        config_file = tmp_path / "config.yaml"
        config_file.write_text(
            dedent(
                """
                model:
                  lr: 0.01
                """
            )
        )

        prov = rc.get_provenance(config_file)

        with pytest.raises(ValueError, match="Unknown preset"):
            rc.format(prov).preset("nonexistent")

    def test_register_provenance_preset__ViaPublicAPI__Works(self) -> None:
        """Test that register_provenance_preset works via rc module."""
        rc.register_provenance_preset(
            "test_preset",
            lambda: rc.ProvenanceFormatContext(show_paths=True),
            "Test preset",
        )

        presets = rc.known_provenance_presets()

        assert "test_preset" in presets
        assert presets["test_preset"].description == "Test preset"
        assert not presets["test_preset"].builtin

    def test_provenance_preset_decorator__RegistersPreset(self) -> None:
        """Test that @provenance_preset decorator registers preset."""

        @rc.provenance_preset("decorated_preset", "Decorated test preset")
        def my_preset() -> rc.ProvenanceFormatContext:
            return rc.ProvenanceFormatContext(show_paths=True, show_values=False)

        presets = rc.known_provenance_presets()

        assert "decorated_preset" in presets
        assert presets["decorated_preset"].description == "Decorated test preset"

        # Verify the factory returns correct context
        ctx = presets["decorated_preset"].factory()
        assert ctx.show_paths is True
        assert ctx.show_values is False

    def test_known_provenance_presets__ReturnsAllPresets(self) -> None:
        """Test that known_provenance_presets returns all presets."""
        presets = rc.known_provenance_presets()

        # Should include all built-in presets
        assert "default" in presets
        assert "minimal" in presets
        assert "compact" in presets
        assert "full" in presets
        assert "values" in presets
        assert "help" in presets
        assert "deprecations" in presets

    def test_known_provenance_presets__IncludesBuiltins(self) -> None:
        """Test that all built-in presets are marked as builtin."""
        presets = rc.known_provenance_presets()

        for name in ["default", "minimal", "compact", "full", "values", "help", "deprecations"]:
            assert presets[name].builtin is True


class TestDiffPresetIntegration:
    """Integration tests for diff preset registration and usage."""

    @pytest.fixture(autouse=True)
    def clear_registries(self):
        """Clear custom presets before and after each test."""
        get_provenance_registry().clear_presets()
        get_diff_registry().clear_presets()
        yield
        get_provenance_registry().clear_presets()
        get_diff_registry().clear_presets()

    def test_DiffFormat__StringPreset__AppliesCorrectly(self, tmp_path: Path) -> None:
        """Test that string-based preset() method applies presets correctly."""
        config_v1 = tmp_path / "v1.yaml"
        config_v1.write_text(
            dedent(
                """
                model:
                  lr: 0.01
                """
            )
        )

        config_v2 = tmp_path / "v2.yaml"
        config_v2.write_text(
            dedent(
                """
                model:
                  lr: 0.02
                  dropout: 0.1
                """
            )
        )

        diff = rc.diff(config_v1, config_v2)

        # Test various presets work
        result_default = str(rc.format(diff).preset("default"))
        result_summary = str(rc.format(diff).preset("summary"))

        # Summary should be shorter (only counts, no entries)
        assert len(result_summary) < len(result_default)

    def test_DiffFormat__NamedMethod__AppliesCorrectly(self, tmp_path: Path) -> None:
        """Test that named preset methods work correctly."""
        config_v1 = tmp_path / "v1.yaml"
        config_v1.write_text("model:\n  lr: 0.01\n")

        config_v2 = tmp_path / "v2.yaml"
        config_v2.write_text("model:\n  lr: 0.02\n")

        diff = rc.diff(config_v1, config_v2)

        # Named methods should produce same output as preset()
        result_changes_method = str(rc.format(diff).changes_only())
        result_changes_preset = str(rc.format(diff).preset("changes_only"))

        assert result_changes_method == result_changes_preset

    def test_DiffFormat__CustomPreset__AppliesCorrectly(self, tmp_path: Path) -> None:
        """Test that custom presets can be registered and used."""
        config_v1 = tmp_path / "v1.yaml"
        config_v1.write_text("model:\n  lr: 0.01\n")

        config_v2 = tmp_path / "v2.yaml"
        config_v2.write_text("model:\n  lr: 0.01\n  new_param: true\n")

        # Register custom preset
        rc.register_diff_preset(
            "added_only",
            lambda: rc.DiffFormatContext(
                show_added=True,
                show_removed=False,
                show_changed=False,
                show_unchanged=False,
            ),
            "Show only added entries",
        )

        diff = rc.diff(config_v1, config_v2)
        result = str(rc.format(diff).preset("added_only"))

        # Should show added entry
        assert "new_param" in result

    def test_DiffFormat__UnknownPreset__RaisesValueError(self, tmp_path: Path) -> None:
        """Test that unknown preset raises ValueError."""
        config_file = tmp_path / "config.yaml"
        config_file.write_text("model:\n  lr: 0.01\n")

        diff = rc.diff(config_file, config_file)

        with pytest.raises(ValueError, match="Unknown preset"):
            rc.format(diff).preset("nonexistent")

    def test_register_diff_preset__ViaPublicAPI__Works(self) -> None:
        """Test that register_diff_preset works via rc module."""
        rc.register_diff_preset(
            "test_diff_preset",
            lambda: rc.DiffFormatContext(show_added=True),
            "Test diff preset",
        )

        presets = rc.known_diff_presets()

        assert "test_diff_preset" in presets
        assert presets["test_diff_preset"].description == "Test diff preset"
        assert not presets["test_diff_preset"].builtin

    def test_diff_preset_decorator__RegistersPreset(self) -> None:
        """Test that @diff_preset decorator registers preset."""

        @rc.diff_preset("decorated_diff_preset", "Decorated diff preset")
        def my_diff_preset() -> rc.DiffFormatContext:
            return rc.DiffFormatContext(show_added=True, show_removed=False)

        presets = rc.known_diff_presets()

        assert "decorated_diff_preset" in presets
        assert presets["decorated_diff_preset"].description == "Decorated diff preset"

        # Verify the factory returns correct context
        ctx = presets["decorated_diff_preset"].factory()
        assert ctx.show_added is True
        assert ctx.show_removed is False

    def test_known_diff_presets__ReturnsAllPresets(self) -> None:
        """Test that known_diff_presets returns all presets."""
        presets = rc.known_diff_presets()

        # Should include all built-in presets
        assert "default" in presets
        assert "changes_only" in presets
        assert "with_context" in presets
        assert "full" in presets
        assert "summary" in presets

    def test_known_diff_presets__IncludesBuiltins(self) -> None:
        """Test that all built-in presets are marked as builtin."""
        presets = rc.known_diff_presets()

        for name in ["default", "changes_only", "with_context", "full", "summary"]:
            assert presets[name].builtin is True


class TestBuiltinProtection:
    """Tests for built-in preset protection."""

    @pytest.fixture(autouse=True)
    def clear_registries(self):
        """Clear custom presets before and after each test."""
        get_provenance_registry().clear_presets()
        get_diff_registry().clear_presets()
        yield
        get_provenance_registry().clear_presets()
        get_diff_registry().clear_presets()

    def test_register_provenance_preset__BuiltinName__RaisesValueError(self) -> None:
        """Test that registering over a built-in preset raises ValueError."""
        with pytest.raises(ValueError, match="built-in preset"):
            rc.register_provenance_preset(
                "minimal",
                lambda: rc.ProvenanceFormatContext(),
                "Override attempt",
            )

    def test_unregister_provenance_preset__BuiltinName__RaisesValueError(self) -> None:
        """Test that unregistering a built-in preset raises ValueError."""
        with pytest.raises(ValueError, match="built-in preset"):
            rc.unregister_provenance_preset("minimal")

    def test_register_diff_preset__BuiltinName__RaisesValueError(self) -> None:
        """Test that registering over a built-in diff preset raises ValueError."""
        with pytest.raises(ValueError, match="built-in preset"):
            rc.register_diff_preset(
                "changes_only",
                lambda: rc.DiffFormatContext(),
                "Override attempt",
            )

    def test_unregister_diff_preset__BuiltinName__RaisesValueError(self) -> None:
        """Test that unregistering a built-in diff preset raises ValueError."""
        with pytest.raises(ValueError, match="built-in preset"):
            rc.unregister_diff_preset("changes_only")


class TestProvenanceLayoutIntegration:
    """Integration tests for provenance layout registration and usage."""

    @pytest.fixture(autouse=True)
    def clear_registries(self):
        """Clear custom layouts before and after each test."""
        get_provenance_registry().clear_layouts()
        get_diff_registry().clear_layouts()
        yield
        get_provenance_registry().clear_layouts()
        get_diff_registry().clear_layouts()

    def test_ProvenanceFormat__StringLayout__AppliesCorrectly(self, tmp_path: Path) -> None:
        """Test that string-based layout() method applies layouts correctly."""
        config_file = tmp_path / "config.yaml"
        config_file.write_text(
            dedent(
                """
                model:
                  lr: 0.01
                  layers: 4
                """
            )
        )

        prov = rc.get_provenance(config_file)

        # Test various layouts work
        result_tree = str(rc.format(prov).layout("tree"))
        result_flat = str(rc.format(prov).layout("flat"))
        result_markdown = str(rc.format(prov).layout("markdown"))

        # All should contain the path
        assert "model.lr" in result_tree
        assert "model.lr" in result_flat
        assert "model.lr" in result_markdown

        # Markdown should have table markers
        assert "|" in result_markdown

    def test_ProvenanceFormat__ConvenienceMethod__AppliesCorrectly(self, tmp_path: Path) -> None:
        """Test that convenience layout methods work correctly."""
        config_file = tmp_path / "config.yaml"
        config_file.write_text(
            dedent(
                """
                model:
                  lr: 0.01
                """
            )
        )

        prov = rc.get_provenance(config_file)

        # Convenience methods should produce same output as layout()
        result_tree_method = str(rc.format(prov).tree())
        result_tree_layout = str(rc.format(prov).layout("tree"))

        assert result_tree_method == result_tree_layout

    def test_ProvenanceFormat__UnknownLayout__RaisesValueError(self, tmp_path: Path) -> None:
        """Test that unknown layout raises ValueError."""
        config_file = tmp_path / "config.yaml"
        config_file.write_text(
            dedent(
                """
                model:
                  lr: 0.01
                """
            )
        )

        prov = rc.get_provenance(config_file)

        with pytest.raises(ValueError, match="Unknown layout"):
            rc.format(prov).layout("nonexistent")

    def test_register_provenance_layout__ViaPublicAPI__Works(self) -> None:
        """Test that register_provenance_layout works via rc module."""
        from rconfig.provenance.formatting import ProvenanceLayout, ProvenanceDisplayModel

        class TestLayout(ProvenanceLayout):
            def render(self, model: ProvenanceDisplayModel) -> str:
                return "test output"

        rc.register_provenance_layout(
            "test_layout",
            lambda: TestLayout(),
            "Test layout",
        )

        layouts = rc.known_provenance_layouts()

        assert "test_layout" in layouts
        assert layouts["test_layout"].description == "Test layout"
        assert not layouts["test_layout"].builtin

    def test_provenance_layout_decorator__RegistersLayout(self) -> None:
        """Test that @provenance_layout decorator registers layout."""
        from rconfig.provenance.formatting import ProvenanceLayout, ProvenanceDisplayModel

        class DecoratedLayout(ProvenanceLayout):
            def render(self, model: ProvenanceDisplayModel) -> str:
                return "decorated output"

        @rc.provenance_layout("decorated_layout", "Decorated test layout")
        def my_layout() -> ProvenanceLayout:
            return DecoratedLayout()

        layouts = rc.known_provenance_layouts()

        assert "decorated_layout" in layouts
        assert layouts["decorated_layout"].description == "Decorated test layout"

        # Verify the factory returns correct layout
        layout = layouts["decorated_layout"].factory()
        assert isinstance(layout, ProvenanceLayout)

    def test_known_provenance_layouts__ReturnsAllLayouts(self) -> None:
        """Test that known_provenance_layouts returns all layouts."""
        layouts = rc.known_provenance_layouts()

        # Should include all built-in layouts
        assert "tree" in layouts
        assert "flat" in layouts
        assert "markdown" in layouts

    def test_known_provenance_layouts__IncludesBuiltins(self) -> None:
        """Test that all built-in layouts are marked as builtin."""
        layouts = rc.known_provenance_layouts()

        for name in ["tree", "flat", "markdown"]:
            assert layouts[name].builtin is True


class TestDiffLayoutIntegration:
    """Integration tests for diff layout registration and usage."""

    @pytest.fixture(autouse=True)
    def clear_registries(self):
        """Clear custom layouts before and after each test."""
        get_provenance_registry().clear_layouts()
        get_diff_registry().clear_layouts()
        yield
        get_provenance_registry().clear_layouts()
        get_diff_registry().clear_layouts()

    def test_DiffFormat__StringLayout__AppliesCorrectly(self, tmp_path: Path) -> None:
        """Test that string-based layout() method applies layouts correctly."""
        config_v1 = tmp_path / "v1.yaml"
        config_v1.write_text(
            dedent(
                """
                model:
                  lr: 0.01
                """
            )
        )

        config_v2 = tmp_path / "v2.yaml"
        config_v2.write_text(
            dedent(
                """
                model:
                  lr: 0.02
                  dropout: 0.1
                """
            )
        )

        diff = rc.diff(config_v1, config_v2)

        # Test various layouts work
        result_flat = str(rc.format(diff).layout("flat"))
        result_tree = str(rc.format(diff).layout("tree"))
        result_markdown = str(rc.format(diff).layout("markdown"))

        # All should contain some indication of changes
        assert "lr" in result_flat or "dropout" in result_flat
        assert "lr" in result_tree or "dropout" in result_tree
        assert "|" in result_markdown  # Table markers

    def test_DiffFormat__ConvenienceMethod__AppliesCorrectly(self, tmp_path: Path) -> None:
        """Test that convenience layout methods work correctly."""
        config_v1 = tmp_path / "v1.yaml"
        config_v1.write_text("model:\n  lr: 0.01\n")

        config_v2 = tmp_path / "v2.yaml"
        config_v2.write_text("model:\n  lr: 0.02\n")

        diff = rc.diff(config_v1, config_v2)

        # tree() should produce same output as layout("tree")
        result_tree_method = str(rc.format(diff).tree())
        result_tree_layout = str(rc.format(diff).layout("tree"))

        assert result_tree_method == result_tree_layout

    def test_DiffFormat__UnknownLayout__RaisesValueError(self, tmp_path: Path) -> None:
        """Test that unknown layout raises ValueError."""
        config_file = tmp_path / "config.yaml"
        config_file.write_text("model:\n  lr: 0.01\n")

        diff = rc.diff(config_file, config_file)

        with pytest.raises(ValueError, match="Unknown layout"):
            rc.format(diff).layout("nonexistent")

    def test_register_diff_layout__ViaPublicAPI__Works(self) -> None:
        """Test that register_diff_layout works via rc module."""
        from rconfig.diff.formatting import DiffLayout, DiffDisplayModel

        class TestDiffLayout(DiffLayout):
            def render(self, model: DiffDisplayModel) -> str:
                return "test diff output"

        rc.register_diff_layout(
            "test_diff_layout",
            lambda: TestDiffLayout(),
            "Test diff layout",
        )

        layouts = rc.known_diff_layouts()

        assert "test_diff_layout" in layouts
        assert layouts["test_diff_layout"].description == "Test diff layout"
        assert not layouts["test_diff_layout"].builtin

    def test_diff_layout_decorator__RegistersLayout(self) -> None:
        """Test that @diff_layout decorator registers layout."""
        from rconfig.diff.formatting import DiffLayout, DiffDisplayModel

        class DecoratedDiffLayout(DiffLayout):
            def render(self, model: DiffDisplayModel) -> str:
                return "decorated diff output"

        @rc.diff_layout("decorated_diff_layout", "Decorated diff layout")
        def my_diff_layout() -> DiffLayout:
            return DecoratedDiffLayout()

        layouts = rc.known_diff_layouts()

        assert "decorated_diff_layout" in layouts
        assert layouts["decorated_diff_layout"].description == "Decorated diff layout"

        # Verify the factory returns correct layout
        layout = layouts["decorated_diff_layout"].factory()
        assert isinstance(layout, DiffLayout)

    def test_known_diff_layouts__ReturnsAllLayouts(self) -> None:
        """Test that known_diff_layouts returns all layouts."""
        layouts = rc.known_diff_layouts()

        # Should include all built-in layouts
        assert "flat" in layouts
        assert "tree" in layouts
        assert "markdown" in layouts

    def test_known_diff_layouts__IncludesBuiltins(self) -> None:
        """Test that all built-in layouts are marked as builtin."""
        layouts = rc.known_diff_layouts()

        for name in ["flat", "tree", "markdown"]:
            assert layouts[name].builtin is True


class TestBuiltinLayoutProtection:
    """Tests for built-in layout protection."""

    @pytest.fixture(autouse=True)
    def clear_registries(self):
        """Clear custom layouts before and after each test."""
        get_provenance_registry().clear_layouts()
        get_diff_registry().clear_layouts()
        yield
        get_provenance_registry().clear_layouts()
        get_diff_registry().clear_layouts()

    def test_register_provenance_layout__BuiltinName__RaisesValueError(self) -> None:
        """Test that registering over a built-in layout raises ValueError."""
        from rconfig.provenance.formatting import ProvenanceTreeLayout

        with pytest.raises(ValueError, match="built-in layout"):
            rc.register_provenance_layout(
                "tree",
                lambda: ProvenanceTreeLayout(),
                "Override attempt",
            )

    def test_unregister_provenance_layout__BuiltinName__RaisesValueError(self) -> None:
        """Test that unregistering a built-in layout raises ValueError."""
        with pytest.raises(ValueError, match="built-in layout"):
            rc.unregister_provenance_layout("tree")

    def test_register_diff_layout__BuiltinName__RaisesValueError(self) -> None:
        """Test that registering over a built-in diff layout raises ValueError."""
        from rconfig.diff.formatting import DiffFlatLayout

        with pytest.raises(ValueError, match="built-in layout"):
            rc.register_diff_layout(
                "flat",
                lambda: DiffFlatLayout(),
                "Override attempt",
            )

    def test_unregister_diff_layout__BuiltinName__RaisesValueError(self) -> None:
        """Test that unregistering a built-in diff layout raises ValueError."""
        with pytest.raises(ValueError, match="built-in layout"):
            rc.unregister_diff_layout("flat")


"""Tree-style layout for diff formatting.

This module provides a grouped tree layout for diffs
that organizes entries by their diff type.
"""

from __future__ import annotations

from rconfig._internal.format_utils import indent

from .model import DiffDisplayModel, DiffEntryDisplayModel
from .layout import DiffLayout
from ..models import DiffEntryType


class DiffTreeLayout(DiffLayout):
    """Grouped tree-style layout for diffs.

    Organizes diff entries by type (Added, Removed, Changed, Unchanged)
    for easier scanning of changes.

    Example output::

        ConfigDiff:
          Added:
            + model.dropout: 0.1
            + training.early_stopping: true

          Removed:
            - model.legacy_param: "old_value"

          Changed:
            ~ model.lr: 0.001 -> 0.0001
            ~ data.batch_size: 32 -> 64

        Added: 2, Removed: 1, Changed: 2

    :param indent_size: Number of spaces per indentation level.
    """

    def __init__(self, indent_size: int = 2) -> None:
        """Initialize the tree layout.

        :param indent_size: Number of spaces per indentation level.
        """
        self._indent_size = indent_size

    def render(self, model: DiffDisplayModel) -> str:
        """Render the diff model to string output.

        :param model: The display model.
        :return: Formatted string representation.
        """
        # Handle empty case
        if model.empty_message:
            return model.empty_message

        # Group entries by type
        added = [e for e in model.entries if e.diff_type == DiffEntryType.ADDED]
        removed = [e for e in model.entries if e.diff_type == DiffEntryType.REMOVED]
        changed = [e for e in model.entries if e.diff_type == DiffEntryType.CHANGED]
        unchanged = [e for e in model.entries if e.diff_type == DiffEntryType.UNCHANGED]

        sections: list[str] = []
        sections.append("ConfigDiff:")

        # Added section
        if added:
            added_section = self._render_section("Added", added)
            if added_section:
                sections.append(added_section)

        # Removed section
        if removed:
            removed_section = self._render_section("Removed", removed)
            if removed_section:
                sections.append(removed_section)

        # Changed section
        if changed:
            changed_section = self._render_section("Changed", changed)
            if changed_section:
                sections.append(changed_section)

        # Unchanged section
        if unchanged:
            unchanged_section = self._render_section("Unchanged", unchanged)
            if unchanged_section:
                sections.append(unchanged_section)

        result = "\n".join(sections)

        # Add summary if present
        if model.summary:
            result = result + "\n\n" + model.summary

        return result

    def _render_section(
        self,
        title: str,
        entries: list[DiffEntryDisplayModel],
    ) -> str:
        """Render a section of entries.

        :param title: Section title (e.g., "Added").
        :param entries: List of entry display models.
        :return: Formatted section string.
        """
        lines: list[str] = []
        lines.append(indent(f"{title}:", 1, self._indent_size))

        for entry in entries:
            formatted = self._render_entry(entry)
            if formatted:
                # Indent entry lines
                for line in formatted.split("\n"):
                    lines.append(indent(line, 2, self._indent_size))

        # Return empty if only header
        if len(lines) == 1:
            return ""

        return "\n".join(lines)

    def _render_entry(self, entry: DiffEntryDisplayModel) -> str:
        """Render a single diff entry.

        :param entry: The entry display model.
        :return: Formatted string for this entry.
        """
        line = self._format_entry_line(entry)

        # Add provenance info if present
        provenance = self._format_provenance(entry)
        if provenance:
            line = line + "\n" + indent(provenance, 1, self._indent_size)

        return line

    def _format_entry_line(self, entry: DiffEntryDisplayModel) -> str:
        """Format the main line for an entry.

        :param entry: The entry display model.
        :return: Formatted main line.
        """
        indicator = entry.diff_type.indicator

        match entry.diff_type:
            case DiffEntryType.ADDED:
                parts = [f"{indicator} {entry.path}"]
                if entry.right_value is not None:
                    parts.append(f": {entry.right_value}")
                return "".join(parts)

            case DiffEntryType.REMOVED:
                parts = [f"{indicator} {entry.path}"]
                if entry.left_value is not None:
                    parts.append(f": {entry.left_value}")
                return "".join(parts)

            case DiffEntryType.CHANGED:
                parts = [f"{indicator} {entry.path}"]
                if entry.left_value is not None and entry.right_value is not None:
                    parts.append(f": {entry.left_value} -> {entry.right_value}")
                return "".join(parts)

            case DiffEntryType.UNCHANGED:
                parts = [f"{indicator} {entry.path}"]
                if entry.left_value is not None:
                    parts.append(f": {entry.left_value}")
                return "".join(parts)

    def _format_provenance(self, entry: DiffEntryDisplayModel) -> str | None:
        """Format provenance information for an entry.

        :param entry: The entry display model.
        :return: Formatted provenance string or None.
        """
        parts: list[str] = []

        if entry.left_provenance:
            loc = f"{entry.left_provenance.file}:{entry.left_provenance.line}"
            parts.append(f"left: {loc}")

        if entry.right_provenance:
            loc = f"{entry.right_provenance.file}:{entry.right_provenance.line}"
            parts.append(f"right: {loc}")

        return ", ".join(parts) if parts else None



"""Tests for TreeLayout implementation."""

from unittest import TestCase

from rconfig.provenance import (
    InstanceRef,
    Provenance,
    ProvenanceBuilder,
    ProvenanceEntry,
    ProvenanceFormatContext,
    ProvenanceFormat,
    TreeLayout,
)
from rconfig.interpolation.evaluator import InterpolationSource


def build_and_render(layout: TreeLayout, provenance: Provenance, ctx: ProvenanceFormatContext) -> str:
    """Helper to build display model via format and render with layout."""
    fmt = ProvenanceFormat(provenance, layout)
    # Apply context settings
    fmt._ctx = ctx
    return str(fmt)


class FormatFilterTests(TestCase):
    """Tests for ProvenanceFormat filter matching."""

    def test_matchesFilters__NoFilters__MatchesEverything(self) -> None:
        """Test that entries match when no filters are set."""
        builder = ProvenanceBuilder()
        builder.add("model.lr", file="config.yaml", line=5, value=42)
        prov = builder.build()

        result = str(ProvenanceFormat(prov))

        self.assertIn("model.lr", result)

    def test_matchesFilters__PathFilterMatch__ReturnsTrue(self) -> None:
        """Test that matching path filter includes entry."""
        builder = ProvenanceBuilder()
        builder.add("model.lr", file="config.yaml", line=5, value=42)
        builder.add("data.path", file="config.yaml", line=6, value="/data")
        prov = builder.build()

        result = str(ProvenanceFormat(prov).for_path("/model.*"))

        self.assertIn("model.lr", result)
        self.assertNotIn("data.path", result)

    def test_matchesFilters__PathFilterNoMatch__ReturnsFalse(self) -> None:
        """Test that non-matching path filter excludes entry."""
        builder = ProvenanceBuilder()
        builder.add("model.lr", file="config.yaml", line=5, value=42)
        prov = builder.build()

        result = str(ProvenanceFormat(prov).for_path("/data.*"))

        self.assertNotIn("model.lr", result)

    def test_matchesFilters__FileFilterMatch__ReturnsTrue(self) -> None:
        """Test that matching file filter includes entry."""
        builder = ProvenanceBuilder()
        builder.add("model.lr", file="config.yaml", line=5, value=42)
        builder.add("data.path", file="config.json", line=6, value="/data")
        prov = builder.build()

        result = str(ProvenanceFormat(prov).from_file("*.yaml"))

        self.assertIn("model.lr", result)
        self.assertNotIn("data.path", result)

    def test_matchesFilters__FileFilterNoMatch__ReturnsFalse(self) -> None:
        """Test that non-matching file filter excludes entry."""
        builder = ProvenanceBuilder()
        builder.add("model.lr", file="config.yaml", line=5, value=42)
        prov = builder.build()

        result = str(ProvenanceFormat(prov).from_file("*.json"))

        self.assertNotIn("model.lr", result)

    def test_matchesFilters__BothFilters__MustMatchBoth(self) -> None:
        """Test that both path and file filters must match (AND logic)."""
        builder = ProvenanceBuilder()
        builder.add("model.lr", file="config.yaml", line=5, value=42)
        prov = builder.build()

        result = str(ProvenanceFormat(prov).for_path("/model.*").from_file("*.json"))

        self.assertNotIn("model.lr", result)

    def test_matchesFilters__MultiplePathFilters__ORLogic(self) -> None:
        """Test that multiple path filters use OR logic."""
        builder = ProvenanceBuilder()
        builder.add("model.lr", file="config.yaml", line=5, value=42)
        builder.add("data.path", file="config.yaml", line=6, value="/data")
        prov = builder.build()

        result = str(ProvenanceFormat(prov).for_path("/data.*").for_path("/model.*"))

        self.assertIn("model.lr", result)
        self.assertIn("data.path", result)

    def test_matchesFilters__MultipleFileFilters__ORLogic(self) -> None:
        """Test that multiple file filters use OR logic."""
        builder = ProvenanceBuilder()
        builder.add("model.lr", file="config.yaml", line=5, value=42)
        builder.add("data.path", file="config.json", line=6, value="/data")
        prov = builder.build()

        result = str(ProvenanceFormat(prov).from_file("*.json").from_file("*.yaml"))

        self.assertIn("model.lr", result)
        self.assertIn("data.path", result)

    def test_matchesFilters__GlobPatternAsterisk__MatchesWildcard(self) -> None:
        """Test that glob wildcard patterns work correctly."""
        builder = ProvenanceBuilder()
        builder.add("model.lr", file="configs/model/base.yaml", line=5, value=42)
        builder.add("other", file="configs/other.yaml", line=6, value=1)
        prov = builder.build()

        result = str(ProvenanceFormat(prov).from_file("configs/model/*.yaml"))

        self.assertIn("model.lr", result)
        self.assertNotIn("other", result)

    def test_matchesFilters__PathWithLeadingSlash__Matches(self) -> None:
        """Test that path matching works with leading slash."""
        builder = ProvenanceBuilder()
        builder.add("model.lr", file="config.yaml", line=5, value=42)
        prov = builder.build()

        result = str(ProvenanceFormat(prov).for_path("/model.lr"))

        self.assertIn("model.lr", result)

    def test_matchesFilters__PathWithoutLeadingSlash__Matches(self) -> None:
        """Test that path matching works without leading slash in filter."""
        builder = ProvenanceBuilder()
        builder.add("model.lr", file="config.yaml", line=5, value=42)
        prov = builder.build()

        result = str(ProvenanceFormat(prov).for_path("model.lr"))

        self.assertIn("model.lr", result)


class TreeLayoutRenderTests(TestCase):
    """Tests for TreeLayout rendering."""

    def setUp(self) -> None:
        self.layout = TreeLayout()

    def test_render__EmptyProvenance__ReturnsEmpty(self) -> None:
        """Test that empty provenance returns empty string."""
        provenance = Provenance()
        ctx = ProvenanceFormatContext()

        result = build_and_render(self.layout, provenance, ctx)

        self.assertEqual("", result)

    def test_render__SingleEntry__FormatsCorrectly(self) -> None:
        """Test single entry is formatted correctly."""
        builder = ProvenanceBuilder()
        builder.add("model.lr", file="config.yaml", line=5, value=0.01)
        provenance = builder.build()
        ctx = ProvenanceFormatContext()

        result = build_and_render(self.layout, provenance, ctx)

        self.assertIn("/model.lr", result)
        self.assertIn("0.01", result)
        self.assertIn("config.yaml:5", result)

    def test_render__MultipleEntries__SeparatesWithBlankLine(self) -> None:
        """Test multiple entries are separated by blank lines."""
        builder = ProvenanceBuilder()
        builder.add("a", file="a.yaml", line=1, value=1)
        builder.add("b", file="b.yaml", line=2, value=2)
        provenance = builder.build()
        ctx = ProvenanceFormatContext()

        result = build_and_render(self.layout, provenance, ctx)

        self.assertIn("\n\n", result)

    def test_render__FilterApplied__FiltersEntries(self) -> None:
        """Test that path filter is applied to entries."""
        builder = ProvenanceBuilder()
        builder.add("model.lr", file="config.yaml", line=1, value=0.01)
        builder.add("data.path", file="config.yaml", line=2, value="/data")
        provenance = builder.build()
        ctx = ProvenanceFormatContext(path_filters=["/model.*"])

        result = build_and_render(self.layout, provenance, ctx)

        self.assertIn("/model.lr", result)
        self.assertNotIn("/data.path", result)

    def test_render__ShowPathsTrue__IncludesPath(self) -> None:
        """Test that path is included when show_paths is True."""
        builder = ProvenanceBuilder()
        builder.add("model.lr", file="config.yaml", line=5, value=42)
        provenance = builder.build()
        ctx = ProvenanceFormatContext(show_paths=True)

        result = build_and_render(self.layout, provenance, ctx)

        self.assertIn("/model.lr", result)

    def test_render__ShowPathsFalse__OmitsPath(self) -> None:
        """Test that path is omitted when show_paths is False."""
        builder = ProvenanceBuilder()
        builder.add("model.lr", file="config.yaml", line=5, value=42)
        provenance = builder.build()
        ctx = ProvenanceFormatContext(show_paths=False)

        result = build_and_render(self.layout, provenance, ctx)

        self.assertNotIn("/model.lr", result)

    def test_render__ShowValuesTrue__IncludesValue(self) -> None:
        """Test that value is included when show_values is True."""
        builder = ProvenanceBuilder()
        builder.add("model.lr", file="config.yaml", line=5, value=42)
        provenance = builder.build()
        ctx = ProvenanceFormatContext(show_values=True)

        result = build_and_render(self.layout, provenance, ctx)

        self.assertIn("42", result)

    def test_render__CliSourceWithArg__ShowsCliAndArg(self) -> None:
        """Test that CLI source shows CLI: and the argument."""
        builder = ProvenanceBuilder()
        builder.add(
            "model.lr",
            file="<override>",
            line=0,
            value=0.01,
            source_type="cli",
            cli_arg="--model.lr=0.01",
        )
        provenance = builder.build()
        ctx = ProvenanceFormatContext(show_source_type=True)

        result = build_and_render(self.layout, provenance, ctx)

        self.assertIn("CLI", result)
        self.assertIn("--model.lr=0.01", result)

    def test_render__EnvSourceWithVar__ShowsEnvAndVar(self) -> None:
        """Test that env source shows env: and the variable name."""
        builder = ProvenanceBuilder()
        builder.add(
            "data.path",
            file="<override>",
            line=0,
            value="/data",
            source_type="env",
            env_var="DATA_PATH",
        )
        provenance = builder.build()
        ctx = ProvenanceFormatContext(show_source_type=True)

        result = build_and_render(self.layout, provenance, ctx)

        self.assertIn("env", result)
        self.assertIn("DATA_PATH", result)

    def test_render__FileSourceWithLocation__ShowsLocation(self) -> None:
        """Test that file source shows file:line location."""
        builder = ProvenanceBuilder()
        builder.add("model.lr", file="config.yaml", line=5, value=42)
        provenance = builder.build()
        ctx = ProvenanceFormatContext(show_files=True, show_lines=True)

        result = build_and_render(self.layout, provenance, ctx)

        self.assertIn("config.yaml:5", result)

    def test_render__WithInterpolation__ShowsInterpolationSection(self) -> None:
        """Test that entry with interpolation shows Interpolation: section."""
        builder = ProvenanceBuilder()
        builder.add(
            "model.lr",
            file="config.yaml",
            line=5,
            value=0.02,
            interpolation=InterpolationSource(
                kind="expression",
                expression="/defaults.lr * 2",
                value=0.02,
                operator="*",
                sources=[
                    InterpolationSource(kind="config", expression="/defaults.lr", value=0.01, path="defaults.lr"),
                    InterpolationSource(kind="literal", expression="2", value=2),
                ],
            ),
        )
        provenance = builder.build()
        ctx = ProvenanceFormatContext(show_chain=True)

        result = build_and_render(self.layout, provenance, ctx)

        self.assertIn("Interpolation:", result)
        self.assertIn("/defaults.lr * 2", result)

    def test_render__WithInstanceChain__ShowsInstances(self) -> None:
        """Test that entry with instance chain shows Instance: lines."""
        builder = ProvenanceBuilder()
        builder.add(
            "service.db",
            file="app.yaml",
            line=10,
            value={"host": "localhost"},
            instance=[
                InstanceRef(path="/shared.database", file="shared.yaml", line=5),
            ],
        )
        provenance = builder.build()
        ctx = ProvenanceFormatContext(show_chain=True)

        result = build_and_render(self.layout, provenance, ctx)

        self.assertIn("Instance:", result)
        self.assertIn("/shared.database", result)
        self.assertIn("shared.yaml:5", result)

    def test_render__WithOverrode__ShowsOverrideSection(self) -> None:
        """Test that entry with overrode shows Overrode: line."""
        builder = ProvenanceBuilder()
        builder.add(
            "model.lr",
            file="override.yaml",
            line=5,
            value=42,
            overrode="base.yaml:10",
        )
        provenance = builder.build()
        ctx = ProvenanceFormatContext(show_overrides=True)

        result = build_and_render(self.layout, provenance, ctx)

        self.assertIn("Overrode:", result)
        self.assertIn("base.yaml:10", result)

    def test_render__HideChain__OmitsInterpolationAndInstance(self) -> None:
        """Test that show_chain=False hides interpolation and instance info."""
        builder = ProvenanceBuilder()
        builder.add(
            "model.lr",
            file="config.yaml",
            line=5,
            value=0.02,
            interpolation=InterpolationSource(
                kind="config",
                expression="/defaults.lr",
                value=0.01,
                path="defaults.lr",
            ),
            instance=[InstanceRef(path="/shared.db", file="shared.yaml", line=1)],
        )
        provenance = builder.build()
        ctx = ProvenanceFormatContext(show_chain=False)

        result = build_and_render(self.layout, provenance, ctx)

        self.assertNotIn("Interpolation:", result)
        self.assertNotIn("Instance:", result)

    def test_render__HideOverrides__OmitsOverrideSection(self) -> None:
        """Test that show_overrides=False hides override info."""
        builder = ProvenanceBuilder()
        builder.add(
            "model.lr",
            file="override.yaml",
            line=5,
            value=42,
            overrode="base.yaml:10",
        )
        provenance = builder.build()
        ctx = ProvenanceFormatContext(show_overrides=False)

        result = build_and_render(self.layout, provenance, ctx)

        self.assertNotIn("Overrode:", result)

    def test_render__AllHidden__ReturnsMinimal(self) -> None:
        """Test that hiding all options returns minimal output."""
        builder = ProvenanceBuilder()
        builder.add("model.lr", file="config.yaml", line=5, value=42)
        provenance = builder.build()
        ctx = ProvenanceFormatContext(
            show_paths=False,
            show_values=False,
            show_files=False,
            show_lines=False,
            show_source_type=False,
            show_chain=False,
            show_overrides=False,
        )

        result = build_and_render(self.layout, provenance, ctx)

        self.assertEqual("", result)


class TreeLayoutProgrammaticSourceTests(TestCase):
    """Tests for programmatic source type formatting."""

    def setUp(self) -> None:
        self.layout = TreeLayout()

    def test_render__ProgrammaticSourceWithoutCliArg__ShowsSourceType(self) -> None:
        """Test that programmatic source without cli_arg shows just source type."""
        builder = ProvenanceBuilder()
        builder.add(
            "test",
            file="<override>",
            line=0,
            value=42,
            source_type="programmatic",
        )
        provenance = builder.build()
        ctx = ProvenanceFormatContext(show_source_type=True)

        result = build_and_render(self.layout, provenance, ctx)

        self.assertIn("programmatic", result)

    def test_render__EnvSourceWithoutEnvVar__ShowsSourceType(self) -> None:
        """Test that env source without env_var shows just source type."""
        builder = ProvenanceBuilder()
        builder.add(
            "test",
            file="<override>",
            line=0,
            value="/data",
            source_type="env",
        )
        provenance = builder.build()
        ctx = ProvenanceFormatContext(show_source_type=True)

        result = build_and_render(self.layout, provenance, ctx)

        self.assertIn("env", result)


class TreeLayoutTargetDisplayTests(TestCase):
    """Tests for TreeLayout target info display."""

    def setUp(self) -> None:
        self.layout = TreeLayout()

    def _build_provenance_with_entry(self, path: str, entry: ProvenanceEntry) -> Provenance:
        """Build a Provenance with a single entry."""
        return Provenance({path: entry})

    def test_render__WithTargetInfo__ShowsTargetLine(self) -> None:
        """Test that target info is displayed."""
        entry = ProvenanceEntry(
            file="config.yaml",
            line=5,
            target_name="model",
            target_class="MyModel",
            target_module="myapp.models",
        )
        provenance = self._build_provenance_with_entry("model", entry)
        ctx = ProvenanceFormatContext(show_targets=True)

        result = build_and_render(self.layout, provenance, ctx)

        self.assertIn("Target: model -> myapp.models.MyModel", result)

    def test_render__WithAutoRegisteredTarget__ShowsAutoRegisteredMarker(self) -> None:
        """Test that auto-registered targets are marked."""
        entry = ProvenanceEntry(
            file="config.yaml",
            line=5,
            target_name="model",
            target_class="MyModel",
            target_module="myapp.models",
            target_auto_registered=True,
        )
        provenance = self._build_provenance_with_entry("model", entry)
        ctx = ProvenanceFormatContext(show_targets=True)

        result = build_and_render(self.layout, provenance, ctx)

        self.assertIn("(auto-registered)", result)

    def test_render__UnregisteredTarget__ShowsNotRegistered(self) -> None:
        """Test that unregistered targets show 'not registered'."""
        entry = ProvenanceEntry(
            file="config.yaml",
            line=5,
            target_name="unknown",
        )
        provenance = self._build_provenance_with_entry("model", entry)
        ctx = ProvenanceFormatContext(show_targets=True)

        result = build_and_render(self.layout, provenance, ctx)

        self.assertIn("Target: unknown (not registered)", result)

    def test_render__ShowTargetsFalse__OmitsTargetLine(self) -> None:
        """Test that targets are hidden when show_targets is False."""
        entry = ProvenanceEntry(
            file="config.yaml",
            line=5,
            target_name="model",
            target_class="MyModel",
            target_module="myapp.models",
        )
        provenance = self._build_provenance_with_entry("model", entry)
        ctx = ProvenanceFormatContext(show_targets=False)

        result = build_and_render(self.layout, provenance, ctx)

        self.assertNotIn("Target:", result)

    def test_render__NoTargetName__OmitsTargetLine(self) -> None:
        """Test that entries without target_name don't show Target line."""
        builder = ProvenanceBuilder()
        builder.add("model.layers", file="config.yaml", line=5)
        provenance = builder.build()
        ctx = ProvenanceFormatContext(show_targets=True)

        result = build_and_render(self.layout, provenance, ctx)

        self.assertNotIn("Target:", result)

    def test_render__ClassWithoutModule__FormatsCorrectly(self) -> None:
        """Test target formatting when class is present but module is None."""
        entry = ProvenanceEntry(
            file="config.yaml",
            line=5,
            target_name="model",
            target_class="MyModel",
        )
        provenance = self._build_provenance_with_entry("model", entry)
        ctx = ProvenanceFormatContext(show_targets=True)

        result = build_and_render(self.layout, provenance, ctx)

        self.assertIn("Target: model -> MyModel", result)


class TreeLayoutDeprecationTests(TestCase):
    """Tests for TreeLayout deprecation formatting."""

    def setUp(self) -> None:
        from rconfig.deprecation.info import DeprecationInfo

        self.layout = TreeLayout()
        builder = ProvenanceBuilder()
        builder.add(
            "old_key",
            file="config.yaml",
            line=5,
            deprecation=DeprecationInfo(
                pattern="old_key",
                new_key="new_key",
                message="Custom deprecation message",
                remove_in="3.0.0",
            ),
        )
        builder.set_config({"old_key": 42})
        self.provenance = builder.build()

    def test_render__WithDeprecation__ShowsDeprecatedMarker(self) -> None:
        ctx = ProvenanceFormatContext()

        result = build_and_render(self.layout, self.provenance, ctx)

        self.assertIn("DEPRECATED", result)

    def test_render__WithNewKey__ShowsNewKeyInArrow(self) -> None:
        ctx = ProvenanceFormatContext()

        result = build_and_render(self.layout, self.provenance, ctx)

        self.assertIn("-> new_key", result)

    def test_render__WithRemoveIn__ShowsVersion(self) -> None:
        ctx = ProvenanceFormatContext()

        result = build_and_render(self.layout, self.provenance, ctx)

        self.assertIn("remove in 3.0.0", result)

    def test_render__WithMessage__ShowsMessage(self) -> None:
        ctx = ProvenanceFormatContext()

        result = build_and_render(self.layout, self.provenance, ctx)

        self.assertIn("Message: Custom deprecation message", result)

    def test_render__HideDeprecations__OmitsDeprecationInfo(self) -> None:
        ctx = ProvenanceFormatContext(show_deprecations=False)

        result = build_and_render(self.layout, self.provenance, ctx)

        self.assertNotIn("DEPRECATED", result)
        self.assertNotIn("new_key", result)

    def test_render__MinimalDeprecation__ShowsOnlyDeprecated(self) -> None:
        from rconfig.deprecation.info import DeprecationInfo

        builder = ProvenanceBuilder()
        builder.add(
            "simple",
            file="config.yaml",
            line=10,
            value="val",
            deprecation=DeprecationInfo(pattern="simple"),
        )
        provenance = builder.build()
        ctx = ProvenanceFormatContext()

        result = build_and_render(self.layout, provenance, ctx)

        self.assertIn("DEPRECATED", result)
        self.assertNotIn("->", result)
        self.assertNotIn("remove in", result)
        self.assertNotIn("Message:", result)

    def test_render__DeprecationsOnly__FiltersCorrectly(self) -> None:
        from rconfig.deprecation.info import DeprecationInfo

        builder = ProvenanceBuilder()
        builder.add(
            "old_key",
            file="config.yaml",
            line=5,
            deprecation=DeprecationInfo(
                pattern="old_key",
                new_key="new_key",
                message="Custom deprecation message",
                remove_in="3.0.0",
            ),
        )
        builder.add("normal", file="config.yaml", line=20)
        builder.set_config({"old_key": 42, "normal": "normal_value"})
        prov = builder.build()
        ctx = ProvenanceFormatContext(deprecations_only=True)

        result = build_and_render(self.layout, prov, ctx)

        self.assertIn("old_key", result)
        self.assertNotIn("normal", result)

    def test_render__DeprecationsOnly__AddsHeader(self) -> None:
        ctx = ProvenanceFormatContext(deprecations_only=True)

        result = build_and_render(self.layout, self.provenance, ctx)

        self.assertIn("Deprecated Keys:", result)

    def test_render__DeprecationsOnlyEmpty__ShowsMessage(self) -> None:
        builder = ProvenanceBuilder()
        builder.add("normal", file="config.yaml", line=1)
        builder.set_config({"normal": 42})
        empty_prov = builder.build()
        ctx = ProvenanceFormatContext(deprecations_only=True)

        result = build_and_render(self.layout, empty_prov, ctx)

        self.assertEqual("No deprecated keys found.", result)


class TreeLayoutFormatValueTests(TestCase):
    """Tests for format_value utility function (used by TreeLayout)."""

    def test_formatValue__LongList__Truncates(self) -> None:
        """Test that long lists are truncated with ellipsis."""
        from rconfig._internal.format_utils import format_value

        long_list = list(range(100))

        result = format_value(long_list)

        if len(str(long_list)) > 50:
            self.assertIn("...", result)
            self.assertLessEqual(len(result), 50)

    def test_formatValue__LongDict__Truncates(self) -> None:
        """Test that long dicts are truncated with ellipsis."""
        from rconfig._internal.format_utils import format_value

        long_dict = {f"key{i}": i for i in range(100)}

        result = format_value(long_dict)

        if len(str(long_dict)) > 50:
            self.assertIn("...", result)

    def test_formatValue__ShortList__NoTruncation(self) -> None:
        """Test that short lists are not truncated."""
        from rconfig._internal.format_utils import format_value

        short_list = [1, 2, 3]

        result = format_value(short_list)

        self.assertEqual("[1, 2, 3]", result)

    def test_formatValue__None__ReturnsNull(self) -> None:
        """Test that None value is formatted as 'null'."""
        from rconfig._internal.format_utils import format_value

        result = format_value(None)

        self.assertEqual(result, "null")

    def test_formatValue__BoolTrue__ReturnsTrue(self) -> None:
        """Test that True is formatted as 'true'."""
        from rconfig._internal.format_utils import format_value

        result = format_value(True)

        self.assertEqual(result, "true")

    def test_formatValue__BoolFalse__ReturnsFalse(self) -> None:
        """Test that False is formatted as 'false'."""
        from rconfig._internal.format_utils import format_value

        result = format_value(False)

        self.assertEqual(result, "false")

    def test_formatValue__String__ReturnsQuotedRepr(self) -> None:
        """Test that strings are formatted with quotes."""
        from rconfig._internal.format_utils import format_value

        result = format_value("hello")

        self.assertEqual(result, "'hello'")

    def test_formatValue__Integer__ReturnsStringified(self) -> None:
        """Test that integers are formatted correctly."""
        from rconfig._internal.format_utils import format_value

        result = format_value(42)

        self.assertEqual(result, "42")



"""Unit tests for hook models."""

from types import MappingProxyType
from unittest import TestCase

from rconfig.hooks.models import HookContext, HookEntry, HookPhase


class HookPhaseTests(TestCase):
    """Tests for the HookPhase enum."""

    def test_HookPhase__AllPhasesExist__EnumComplete(self):
        # Assert
        self.assertIn(HookPhase.CONFIG_LOADED, HookPhase)
        self.assertIn(HookPhase.BEFORE_INSTANTIATE, HookPhase)
        self.assertIn(HookPhase.AFTER_INSTANTIATE, HookPhase)
        self.assertIn(HookPhase.ON_ERROR, HookPhase)
        self.assertEqual(len(HookPhase), 4)

    def test_HookPhase__Name__ReturnsExpectedString(self):
        # Assert
        self.assertEqual(HookPhase.CONFIG_LOADED.name, "CONFIG_LOADED")
        self.assertEqual(HookPhase.BEFORE_INSTANTIATE.name, "BEFORE_INSTANTIATE")
        self.assertEqual(HookPhase.AFTER_INSTANTIATE.name, "AFTER_INSTANTIATE")
        self.assertEqual(HookPhase.ON_ERROR.name, "ON_ERROR")


class HookEntryTests(TestCase):
    """Tests for the HookEntry dataclass."""

    def test_init__AllFields__CreatesEntry(self):
        # Arrange
        def my_hook(ctx: HookContext) -> None:
            pass

        # Act
        entry = HookEntry(
            name="test_hook",
            phase=HookPhase.CONFIG_LOADED,
            func=my_hook,
            pattern="**/model/*.yaml",
            priority=10,
        )

        # Assert
        self.assertEqual(entry.name, "test_hook")
        self.assertEqual(entry.phase, HookPhase.CONFIG_LOADED)
        self.assertEqual(entry.func, my_hook)
        self.assertEqual(entry.pattern, "**/model/*.yaml")
        self.assertEqual(entry.priority, 10)

    def test_init__DefaultValues__UsesDefaults(self):
        # Arrange
        def my_hook(ctx: HookContext) -> None:
            pass

        # Act
        entry = HookEntry(
            name="test_hook",
            phase=HookPhase.CONFIG_LOADED,
            func=my_hook,
        )

        # Assert
        self.assertIsNone(entry.pattern)
        self.assertEqual(entry.priority, 50)

    def test_frozen__Immutable__CannotModify(self):
        # Arrange
        def my_hook(ctx: HookContext) -> None:
            pass

        entry = HookEntry(
            name="test_hook",
            phase=HookPhase.CONFIG_LOADED,
            func=my_hook,
        )

        # Act & Assert
        with self.assertRaises(Exception):  # FrozenInstanceError
            entry.name = "modified"

    def test_frozen__Immutable__CannotModifyPriority(self):
        # Arrange
        def my_hook(ctx: HookContext) -> None:
            pass

        entry = HookEntry(
            name="test_hook",
            phase=HookPhase.CONFIG_LOADED,
            func=my_hook,
        )

        # Act & Assert
        with self.assertRaises(Exception):  # FrozenInstanceError
            entry.priority = 100


class HookContextTests(TestCase):
    """Tests for the HookContext dataclass."""

    def test_init__ConfigLoadedPhase__CreatesContext(self):
        # Arrange
        config = MappingProxyType({"key": "value"})

        # Act
        ctx = HookContext(
            phase=HookPhase.CONFIG_LOADED,
            config_path="config.yaml",
            config=config,
        )

        # Assert
        self.assertEqual(ctx.phase, HookPhase.CONFIG_LOADED)
        self.assertEqual(ctx.config_path, "config.yaml")
        self.assertEqual(ctx.config, config)
        self.assertIsNone(ctx.inner_path)
        self.assertIsNone(ctx.instance)
        self.assertIsNone(ctx.target_name)
        self.assertIsNone(ctx.error)

    def test_init__BeforeInstantiatePhase__CreatesContext(self):
        # Arrange
        config = MappingProxyType({"_target_": "model"})

        # Act
        ctx = HookContext(
            phase=HookPhase.BEFORE_INSTANTIATE,
            config_path="config.yaml",
            config=config,
            inner_path="model.encoder",
            target_name="encoder",
        )

        # Assert
        self.assertEqual(ctx.phase, HookPhase.BEFORE_INSTANTIATE)
        self.assertEqual(ctx.inner_path, "model.encoder")
        self.assertEqual(ctx.target_name, "encoder")

    def test_init__AfterInstantiatePhase__HasInstance(self):
        # Arrange
        instance = object()
        config = MappingProxyType({"_target_": "model"})

        # Act
        ctx = HookContext(
            phase=HookPhase.AFTER_INSTANTIATE,
            config_path="config.yaml",
            config=config,
            inner_path="model",
            target_name="model",
            instance=instance,
        )

        # Assert
        self.assertEqual(ctx.phase, HookPhase.AFTER_INSTANTIATE)
        self.assertIs(ctx.instance, instance)

    def test_init__OnErrorPhase__HasError(self):
        # Arrange
        error = ValueError("test error")
        config = MappingProxyType({"_target_": "model"})

        # Act
        ctx = HookContext(
            phase=HookPhase.ON_ERROR,
            config_path="config.yaml",
            config=config,
            error=error,
        )

        # Assert
        self.assertEqual(ctx.phase, HookPhase.ON_ERROR)
        self.assertIs(ctx.error, error)

    def test_frozen__Immutable__CannotModify(self):
        # Arrange
        ctx = HookContext(
            phase=HookPhase.CONFIG_LOADED,
            config_path="config.yaml",
        )

        # Act & Assert
        with self.assertRaises(Exception):  # FrozenInstanceError
            ctx.config_path = "modified.yaml"

    def test_frozen__Immutable__CannotModifyPhase(self):
        # Arrange
        ctx = HookContext(
            phase=HookPhase.CONFIG_LOADED,
            config_path="config.yaml",
        )

        # Act & Assert
        with self.assertRaises(Exception):  # FrozenInstanceError
            ctx.phase = HookPhase.ON_ERROR



"""MultirunIterator class for lazy iteration over multirun results.

This module defines the iterator type returned by instantiate_multirun(),
providing lazy instantiation with length, slicing, and reversal support.
"""

from __future__ import annotations

from typing import Any, Callable, Generic, TypeVar, overload

from .result import MultirunResult

T = TypeVar("T")


class MultirunIterator(Generic[T]):
    """Lazy iterator with length and slicing support for progress tracking.

    Enables efficient iteration over multirun configurations with support for:
    - `len()` for progress tracking (e.g., with tqdm)
    - Slicing for subset iteration (e.g., `iterator[50:]` to resume)
    - Reversal for reverse order iteration
    - Single index access for debugging specific runs

    Example::

        results = rc.instantiate_multirun(...)

        # Progress tracking with tqdm (auto-detects __len__)
        for result in tqdm(results):
            train(result.instance)

        # Resume from crash
        for result in results[50:]:
            train(result.instance)

        # Distribute across workers
        worker1_results = results[0:25]
        worker2_results = results[25:50]

        # Run in reverse order
        for result in reversed(results):
            train(result.instance)

        # Debug specific run
        single_result = results[3]
    """

    def __init__(
        self,
        run_configs: list[dict[str, Any]],
        instantiate_fn: Callable[[dict[str, Any]], MultirunResult[T]],
    ) -> None:
        """Initialize the multirun iterator.

        :param run_configs: List of override dicts for each run.
        :param instantiate_fn: Function that instantiates a single run config.
        """
        self._run_configs = run_configs
        self._instantiate_fn = instantiate_fn
        self._index = 0

    def __len__(self) -> int:
        """Return the total number of runs.

        :return: Number of run configurations.
        """
        return len(self._run_configs)

    def __iter__(self) -> "MultirunIterator[T]":
        """Return a fresh iterator.

        :return: New iterator starting from the beginning.
        """
        return MultirunIterator(self._run_configs, self._instantiate_fn)

    def __next__(self) -> MultirunResult[T]:
        """Yield the next result.

        :return: MultirunResult for the next run.
        :raises StopIteration: When all runs have been iterated.
        """
        if self._index >= len(self._run_configs):
            raise StopIteration
        result = self._instantiate_fn(self._run_configs[self._index])
        self._index += 1
        return result

    def __reversed__(self) -> "MultirunIterator[T]":
        """Return an iterator in reverse order.

        :return: New iterator with reversed run order.
        """
        return MultirunIterator(
            list(reversed(self._run_configs)),
            self._instantiate_fn,
        )

    @overload
    def __getitem__(self, key: int) -> MultirunResult[T]: ...

    @overload
    def __getitem__(self, key: slice) -> MultirunIterator[T]: ...

    def __getitem__(self, key: int | slice) -> MultirunResult[T] | MultirunIterator[T]:
        """Access runs by index or slice.

        :param key: Integer index or slice.
        :return: MultirunResult for single index, MultirunIterator for slice.
        :raises TypeError: If key is not an int or slice.
        :raises IndexError: If integer index is out of range.
        """
        if isinstance(key, int):
            if key < 0:
                key = len(self) + key
            if key < 0 or key >= len(self):
                raise IndexError(f"Index {key} out of range for {len(self)} runs")
            return self._instantiate_fn(self._run_configs[key])
        elif isinstance(key, slice):
            return MultirunIterator(self._run_configs[key], self._instantiate_fn)
        else:
            raise TypeError(
                f"Indices must be integers or slices, not {type(key).__name__}"
            )



"""Argparse help integration for CLI help generation.

This module provides ArgparseHelpIntegration which integrates config entries
into argparse help output via the epilog.
"""

from __future__ import annotations

import argparse
from typing import TYPE_CHECKING

from .integration import HelpIntegration
from .multirun_help import MULTIRUN_HELP

if TYPE_CHECKING:
    from rconfig.provenance import Provenance


class ArgparseHelpIntegration(HelpIntegration):
    """Integrates config entries into argparse help output.

    This adds config entries to the argparse parser's epilog,
    so they appear when the user runs --help.

    Uses consume_help_flag=False because argparse handles --help itself.

    Example::

        parser = argparse.ArgumentParser()
        parser.add_argument("--verbose", action="store_true")

        rc.set_help_integration(ArgparseHelpIntegration(parser))
        args = parser.parse_args()
        config = rc.instantiate(Path("config.yaml"))

        # --help shows argparse help WITH config entries
    """

    def __init__(self, parser: argparse.ArgumentParser) -> None:
        """Initialize the argparse help integration.

        :param parser: The argparse parser to integrate with.
        """
        super().__init__(consume_help_flag=False)
        self._parser = parser

    def integrate(self, provenance: Provenance, config_path: str) -> None:
        """Add config entries to parser's epilog.

        :param provenance: Provenance data with type hints and descriptions.
        :param config_path: Path to the config file.
        """
        epilog = self._formatted_config_entries(provenance, config_path)
        if self._parser.epilog:
            self._parser.epilog += "\n\n" + epilog
        else:
            self._parser.epilog = epilog

        # Set formatter class to preserve formatting
        self._parser.formatter_class = argparse.RawDescriptionHelpFormatter

    def _formatted_config_entries(
        self, provenance: Provenance, config_path: str
    ) -> str:
        """Format config entries for epilog.

        :param provenance: Provenance data.
        :param config_path: Path to the config file.
        :return: Formatted epilog string.
        """
        lines: list[str] = []

        # Header
        lines.append(f"Configuration options ({config_path}):")
        lines.append("-" * 40)

        # Collect entries for alignment
        entries: list[tuple[str, str, str, str]] = []
        for path, entry in provenance.items():
            type_str = self._formatted_type(entry.type_hint)
            value_str = self._formatted_value(entry.value)
            desc_str = entry.description or ""
            entries.append((path, type_str, value_str, desc_str))

        if not entries:
            lines.append("  No configuration options found.")
            return "\n".join(lines)

        # Calculate column widths
        path_width = max(len(e[0]) for e in entries)
        type_width = max(len(e[1]) for e in entries) if entries else 0
        value_width = max(len(e[2]) for e in entries) if entries else 0

        # Format entries
        for path, type_str, value_str, desc_str in entries:
            line = f"  {path:<{path_width}}  {type_str:<{type_width}}  {value_str:<{value_width}}"
            if desc_str:
                line += f"  {desc_str}"
            lines.append(line)

        lines.append("")
        lines.append("Override with: python script.py key=value")
        lines.append(MULTIRUN_HELP)

        return "\n".join(lines)

    def _formatted_type(self, type_hint: type | None) -> str:
        """Format a type hint for display.

        :param type_hint: The type hint.
        :return: Formatted type string.
        """
        if type_hint is None:
            return ""
        # For generic types (list[int], dict[str, Any], etc.), use str()
        # which gives the full parameterized type representation
        if hasattr(type_hint, "__origin__"):
            return str(type_hint)
        if hasattr(type_hint, "__name__"):
            return type_hint.__name__
        return str(type_hint)

    def _formatted_value(self, value: object) -> str:
        """Format a value for display.

        :param value: The value to format.
        :return: Formatted value string.
        """
        if value is None:
            return "(required)"
        elif isinstance(value, bool):
            return "true" if value else "false"
        elif isinstance(value, str):
            if len(value) > 20:
                return repr(value[:17] + "...")
            return repr(value)
        elif isinstance(value, (list, dict)):
            s = str(value)
            if len(s) > 20:
                return s[:17] + "..."
            return s
        else:
            return str(value)



"""Tests for partial config extraction utilities."""

from unittest import TestCase

from rconfig._internal.partial import (
    collect_external_targets,
    extract_partial_config,
    process_instance_targets,
)
from rconfig.errors import InvalidInnerPathError


class ExtractPartialConfigTests(TestCase):
    """Tests for extract_partial_config function."""

    def test_extract_partial_config__ValidPath__ReturnsSubConfig(self):
        # Arrange
        config = {
            "_target_": "trainer",
            "model": {"_target_": "model", "hidden_size": 256},
            "epochs": 10,
        }
        instance_targets: dict[str, str | None] = {}

        # Act
        sub_config, processed_targets, external_targets = extract_partial_config(
            config=config,
            inner_path="model",
            instance_targets=instance_targets,
        )

        # Assert
        self.assertEqual(sub_config, {"_target_": "model", "hidden_size": 256})
        self.assertEqual(processed_targets, {})
        self.assertEqual(external_targets, set())

    def test_extract_partial_config__NestedPath__ReturnsDeepSubConfig(self):
        # Arrange
        config = {
            "_target_": "trainer",
            "model": {
                "_target_": "model",
                "encoder": {"_target_": "encoder", "layers": 6},
            },
        }
        instance_targets: dict[str, str | None] = {}

        # Act
        sub_config, processed_targets, external_targets = extract_partial_config(
            config=config,
            inner_path="model.encoder",
            instance_targets=instance_targets,
        )

        # Assert
        self.assertEqual(sub_config, {"_target_": "encoder", "layers": 6})

    def test_extract_partial_config__ListIndexPath__ReturnsListElement(self):
        # Arrange
        config = {
            "_target_": "trainer",
            "callbacks": [
                {"_target_": "callback", "name": "first"},
                {"_target_": "callback", "name": "second"},
            ],
        }
        instance_targets: dict[str, str | None] = {}

        # Act
        sub_config, processed_targets, external_targets = extract_partial_config(
            config=config,
            inner_path="callbacks[0]",
            instance_targets=instance_targets,
        )

        # Assert
        self.assertEqual(sub_config, {"_target_": "callback", "name": "first"})

    def test_extract_partial_config__InvalidPath__RaisesInvalidInnerPathError(self):
        # Arrange
        config = {"_target_": "model", "size": 256}
        instance_targets: dict[str, str | None] = {}

        # Act & Assert
        with self.assertRaises(InvalidInnerPathError) as ctx:
            extract_partial_config(
                config=config,
                inner_path="nonexistent",
                instance_targets=instance_targets,
            )

        self.assertEqual(ctx.exception.inner_path, "nonexistent")

    def test_extract_partial_config__PathToNonDict__RaisesInvalidInnerPathError(self):
        # Arrange
        config = {"_target_": "model", "size": 256}
        instance_targets: dict[str, str | None] = {}

        # Act & Assert
        with self.assertRaises(InvalidInnerPathError) as ctx:
            extract_partial_config(
                config=config,
                inner_path="size",
                instance_targets=instance_targets,
            )

        self.assertEqual(ctx.exception.inner_path, "size")
        self.assertIn("int", ctx.exception.reason)

    def test_extract_partial_config__DeepCopiesSubConfig(self):
        # Arrange
        config = {
            "_target_": "trainer",
            "model": {"_target_": "model", "nested": {"value": 42}},
        }
        instance_targets: dict[str, str | None] = {}

        # Act
        sub_config, _, _ = extract_partial_config(
            config=config,
            inner_path="model",
            instance_targets=instance_targets,
        )

        # Modify the extracted config
        sub_config["nested"]["value"] = 100

        # Assert - original should be unchanged
        self.assertEqual(config["model"]["nested"]["value"], 42)


class ProcessInstanceTargetsTests(TestCase):
    """Tests for process_instance_targets function."""

    def test_process_instance_targets__TargetWithinScope__RebasesPath(self):
        # Arrange
        original_targets = {
            "model.encoder": "model.shared_layer",  # Both within "model" scope
        }
        inner_path = "model"

        # Act
        processed, external = process_instance_targets(
            original_targets=original_targets,
            inner_path=inner_path,
        )

        # Assert - paths should be rebased relative to "model"
        self.assertEqual(processed, {"encoder": "shared_layer"})
        self.assertEqual(external, set())

    def test_process_instance_targets__TargetOutsideScope__MarksExternal(self):
        # Arrange
        original_targets = {
            "services.api.db": "shared_db",  # Target is outside "services.api"
        }
        inner_path = "services.api"

        # Act
        processed, external = process_instance_targets(
            original_targets=original_targets,
            inner_path=inner_path,
        )

        # Assert
        self.assertEqual(processed, {"db": "__external__:shared_db"})
        self.assertEqual(external, {"shared_db"})

    def test_process_instance_targets__InstanceOutsideScope__Ignored(self):
        # Arrange
        original_targets = {
            "trainer.model": "shared_model",  # Outside "services" scope
            "services.db": "shared_db",  # Inside "services" scope
        }
        inner_path = "services"

        # Act
        processed, external = process_instance_targets(
            original_targets=original_targets,
            inner_path=inner_path,
        )

        # Assert - only "services.db" is processed, "trainer.model" is ignored
        self.assertEqual(processed, {"db": "__external__:shared_db"})
        self.assertEqual(external, {"shared_db"})

    def test_process_instance_targets__NullTarget__PreservedAsNull(self):
        # Arrange
        original_targets = {
            "model.optional": None,  # _instance_: null
        }
        inner_path = "model"

        # Act
        processed, external = process_instance_targets(
            original_targets=original_targets,
            inner_path=inner_path,
        )

        # Assert
        self.assertEqual(processed, {"optional": None})
        self.assertEqual(external, set())

    def test_process_instance_targets__ExactInnerPath__RebasesToEmpty(self):
        # Arrange
        original_targets = {
            "model": "shared_model",  # Instance IS the inner_path
        }
        inner_path = "model"

        # Act
        processed, external = process_instance_targets(
            original_targets=original_targets,
            inner_path=inner_path,
        )

        # Assert - rebased to empty string (the root of the partial)
        self.assertEqual(processed, {"": "__external__:shared_model"})
        self.assertEqual(external, {"shared_model"})

    def test_process_instance_targets__MultipleInstancesPointingSameExternal__DeduplicatesExternals(
        self,
    ):
        # Arrange
        original_targets = {
            "services.a.cache": "shared_cache",
            "services.b.cache": "shared_cache",  # Same external target
        }
        inner_path = "services"

        # Act
        processed, external = process_instance_targets(
            original_targets=original_targets,
            inner_path=inner_path,
        )

        # Assert
        self.assertEqual(
            processed,
            {
                "a.cache": "__external__:shared_cache",
                "b.cache": "__external__:shared_cache",
            },
        )
        # External targets are deduplicated in the set
        self.assertEqual(external, {"shared_cache"})


class CollectExternalTargetsTests(TestCase):
    """Tests for collect_external_targets function."""

    def test_collect_external_targets__MultipleExternal__ReturnsUniquePaths(self):
        # Arrange
        instance_targets = {
            "a.cache": "__external__:/shared/cache",
            "b.cache": "__external__:/shared/cache",  # Same target
            "c.db": "__external__:/shared/db",  # Different target
            "d.internal": "local",  # Not external
        }

        # Act
        result = collect_external_targets(instance_targets)

        # Assert
        self.assertEqual(result, {"/shared/cache", "/shared/db"})

    def test_collect_external_targets__NoExternals__ReturnsEmptySet(self):
        # Arrange
        instance_targets = {
            "a.ref": "b.target",
            "c.ref": None,
        }

        # Act
        result = collect_external_targets(instance_targets)

        # Assert
        self.assertEqual(result, set())

    def test_collect_external_targets__EmptyTargets__ReturnsEmptySet(self):
        # Arrange
        instance_targets: dict[str, str | None] = {}

        # Act
        result = collect_external_targets(instance_targets)

        # Assert
        self.assertEqual(result, set())
